# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Combine 3 images to produce a properly-scaled RGB image following
Lupton et al. (2004).

The three images must be aligned and have the same pixel scale and size.

For details, see : https://ui.adsabs.harvard.edu/abs/2004PASP..116..133L
"""

import numpy as np

from astropy.visualization.interval import ManualInterval, ZScaleInterval
from astropy.visualization.stretch import BaseStretch
from astropy.visualization.stretch import _prepare as _stretch_prepare

from .basic_rgb import RGBImageMapping

__all__ = [
    "AsinhMapping",
    "AsinhZScaleMapping",
    "LinearMapping",
    "LuptonAsinhStretch",
    "LuptonAsinhZscaleStretch",
    "Mapping",
    "make_lupton_rgb",
]


def compute_intensity(image_r, image_g=None, image_b=None):
    """
    Return a naive total intensity from the red, blue, and green intensities.

    Parameters
    ----------
    image_r : ndarray
        Intensity of image to be mapped to red; or total intensity if
        ``image_g`` and ``image_b`` are None.
    image_g : ndarray, optional
        Intensity of image to be mapped to green.
    image_b : ndarray, optional
        Intensity of image to be mapped to blue.

    Returns
    -------
    intensity : ndarray
        Total intensity from the red, blue and green intensities, or
        ``image_r`` if green and blue images are not provided.

    """
    if image_g is None or image_b is None:
        if not (image_g is None and image_b is None):
            raise ValueError(
                "please specify either a single image or red, green, and blue images."
            )
        return image_r

    intensity = (image_r + image_g + image_b) / 3.0

    # Repack into whatever type was passed to us
    return np.asarray(intensity, dtype=image_r.dtype)


class Mapping:
    """
    Baseclass to map red, blue, green intensities into uint8 values.

    Parameters
    ----------
    minimum : float or sequence(3)
        Intensity that should be mapped to black (a scalar or array for R, G, B).
    image : ndarray, optional
        An image used to calculate some parameters of some mappings.
    """

    def __init__(self, minimum=None, image=None):
        self._uint8Max = float(np.iinfo(np.uint8).max)

        try:
            len(minimum)
        except TypeError:
            minimum = 3 * [minimum]
        if len(minimum) != 3:
            raise ValueError("please provide 1 or 3 values for minimum.")

        self.minimum = minimum
        self._image = np.asarray(image)

    def make_rgb_image(self, image_r, image_g, image_b):
        """
        Convert 3 arrays, image_r, image_g, and image_b into an 8-bit RGB image.

        Parameters
        ----------
        image_r : ndarray
            Image to map to red.
        image_g : ndarray
            Image to map to green.
        image_b : ndarray
            Image to map to blue.

        Returns
        -------
        RGBimage : ndarray
            RGB (integer, 8-bits per channel) color image as an NxNx3 numpy array.
        """
        image_r = np.asarray(image_r)
        image_g = np.asarray(image_g)
        image_b = np.asarray(image_b)

        if (image_r.shape != image_g.shape) or (image_g.shape != image_b.shape):
            msg = "The image shapes must match. r: {}, g: {} b: {}"
            raise ValueError(msg.format(image_r.shape, image_g.shape, image_b.shape))

        return np.dstack(
            self._convert_images_to_uint8(image_r, image_g, image_b)
        ).astype(np.uint8)

    def intensity(self, image_r, image_g, image_b):
        """
        Return the total intensity from the red, blue, and green intensities.
        This is a naive computation, and may be overridden by subclasses.

        Parameters
        ----------
        image_r : ndarray
            Intensity of image to be mapped to red; or total intensity if
            ``image_g`` and ``image_b`` are None.
        image_g : ndarray, optional
            Intensity of image to be mapped to green.
        image_b : ndarray, optional
            Intensity of image to be mapped to blue.

        Returns
        -------
        intensity : ndarray
            Total intensity from the red, blue and green intensities, or
            ``image_r`` if green and blue images are not provided.
        """
        return compute_intensity(image_r, image_g, image_b)

    def map_intensity_to_uint8(self, I):
        """
        Return an array which, when multiplied by an image, returns that image
        mapped to the range of a uint8, [0, 255] (but not converted to uint8).

        The intensity is assumed to have had minimum subtracted (as that can be
        done per-band).

        Parameters
        ----------
        I : ndarray
            Intensity to be mapped.

        Returns
        -------
        mapped_I : ndarray
            ``I`` mapped to uint8
        """
        with np.errstate(invalid="ignore", divide="ignore"):
            return np.clip(I, 0, self._uint8Max)

    def _convert_images_to_uint8(self, image_r, image_g, image_b):
        """
        Use the mapping to convert images image_r, image_g, and image_b to a triplet of uint8 images.
        """
        image_r = image_r - self.minimum[0]  # n.b. makes copy
        image_g = image_g - self.minimum[1]
        image_b = image_b - self.minimum[2]

        fac = self.map_intensity_to_uint8(self.intensity(image_r, image_g, image_b))

        image_rgb = [image_r, image_g, image_b]
        for c in image_rgb:
            c *= fac
            with np.errstate(invalid="ignore"):
                c[c < 0] = 0  # individual bands can still be < 0, even if fac isn't

        pixmax = self._uint8Max
        # copies -- could work row by row to minimise memory usage
        r0, g0, b0 = image_rgb

        # n.b. np.where can't and doesn't short-circuit
        with np.errstate(invalid="ignore", divide="ignore"):
            for i, c in enumerate(image_rgb):
                c = np.where(
                    r0 > g0,
                    np.where(
                        r0 > b0,
                        np.where(r0 >= pixmax, c * pixmax / r0, c),
                        np.where(b0 >= pixmax, c * pixmax / b0, c),
                    ),
                    np.where(
                        g0 > b0,
                        np.where(g0 >= pixmax, c * pixmax / g0, c),
                        np.where(b0 >= pixmax, c * pixmax / b0, c),
                    ),
                ).astype(np.uint8)
                c[c > pixmax] = pixmax

                image_rgb[i] = c

        return image_rgb


class LinearMapping(Mapping):
    """
    A linear map map of red, blue, green intensities into uint8 values.

    A linear stretch from [minimum, maximum].
    If one or both are omitted use image min and/or max to set them.

    Parameters
    ----------
    minimum : float
        Intensity that should be mapped to black (a scalar or array for R, G, B).
    maximum : float
        Intensity that should be mapped to white (a scalar).
    """

    def __init__(self, minimum=None, maximum=None, image=None):
        if minimum is None or maximum is None:
            if image is None:
                raise ValueError(
                    "you must provide an image if you don't "
                    "set both minimum and maximum"
                )
            if minimum is None:
                minimum = image.min()
            if maximum is None:
                maximum = image.max()

        Mapping.__init__(self, minimum=minimum, image=image)
        self.maximum = maximum

        if maximum is None:
            self._range = None
        else:
            if maximum == minimum:
                raise ValueError("minimum and maximum values must not be equal")
            self._range = float(maximum - minimum)

    def map_intensity_to_uint8(self, I):
        # n.b. np.where can't and doesn't short-circuit
        with np.errstate(invalid="ignore", divide="ignore"):
            return np.where(
                I <= 0,
                0,
                np.where(
                    I >= self._range, self._uint8Max / I, self._uint8Max / self._range
                ),
            )


class AsinhMapping(Mapping):
    """
    A mapping for an asinh stretch (preserving colours independent of brightness).

    x = asinh(Q (I - minimum)/stretch)/Q

    This reduces to a linear stretch if Q == 0

    See https://ui.adsabs.harvard.edu/abs/2004PASP..116..133L

    Parameters
    ----------
    minimum : float
        Intensity that should be mapped to black (a scalar or array for R, G, B).
    stretch : float
        The linear stretch of the image.
    Q : float
        The asinh softening parameter.
    """

    def __init__(self, minimum, stretch, Q=8):
        Mapping.__init__(self, minimum)

        # 32bit floating point machine epsilon; sys.float_info.epsilon is 64bit
        epsilon = 1.0 / 2**23
        if abs(Q) < epsilon:
            Q = 0.1
        else:
            Qmax = 1e10
            if Q > Qmax:
                Q = Qmax

        frac = 0.1  # gradient estimated using frac*stretch is _slope
        self._slope = frac * self._uint8Max / np.arcsinh(frac * Q)

        self._soften = Q / float(stretch)

    def map_intensity_to_uint8(self, I):
        # n.b. np.where can't and doesn't short-circuit
        with np.errstate(invalid="ignore", divide="ignore"):
            return np.where(I <= 0, 0, np.arcsinh(I * self._soften) * self._slope / I)


class AsinhZScaleMapping(AsinhMapping):
    """
    A mapping for an asinh stretch, estimating the linear stretch by zscale.

    x = asinh(Q (I - z1)/(z2 - z1))/Q

    Parameters
    ----------
    image1 : ndarray or a list of arrays
        The image to analyse, or a list of 3 images to be converted to
        an intensity image.
    image2 : ndarray, optional
        the second image to analyse (must be specified with image3).
    image3 : ndarray, optional
        the third image to analyse (must be specified with image2).
    Q : float, optional
        The asinh softening parameter. Default is 8.
    pedestal : float or sequence(3), optional
        The value, or array of 3 values, to subtract from the images; or None.

    Notes
    -----
    pedestal, if not None, is removed from the images when calculating the
    zscale stretch, and added back into Mapping.minimum[]
    """

    def __init__(self, image1, image2=None, image3=None, Q=8, pedestal=None):
        if image2 is None or image3 is None:
            if not (image2 is None and image3 is None):
                raise ValueError(
                    "please specify either a single image or three images."
                )
            image = [image1]
        else:
            image = [image1, image2, image3]

        if pedestal is not None:
            try:
                len(pedestal)
            except TypeError:
                pedestal = 3 * [pedestal]

            if len(pedestal) != 3:
                raise ValueError("please provide 1 or 3 pedestals.")

            image = list(image)  # needs to be mutable
            for i, im in enumerate(image):
                if pedestal[i] != 0.0:
                    image[i] = im - pedestal[i]  # n.b. a copy
        else:
            pedestal = len(image) * [0.0]

        image = compute_intensity(*image)

        zscale_limits = ZScaleInterval().get_limits(image)
        zscale = LinearMapping(*zscale_limits, image=image)
        # zscale.minimum is always a triple
        stretch = zscale.maximum - zscale.minimum[0]
        minimum = zscale.minimum

        for i, level in enumerate(pedestal):
            minimum[i] += level

        AsinhMapping.__init__(self, minimum, stretch, Q)
        self._image = image


class LuptonAsinhStretch(BaseStretch):
    r"""
    A modified asinh stretch, with some changes to the constants
    relative to `~astropy.visualization.AsinhStretch`.

    The stretch is given by:

    .. math::
        & y = {\rm asinh}\left(\frac{Q * x}{stretch}\right) *
        \frac{frac}{{\rm asinh}(frac * Q)} \\
        & frac = 0.1

    Parameters
    ----------
    stretch : float, optional
        Linear stretch of the image. ``stretch`` must be greater than 0.
        Default is 5.

    Q : float, optional
        The asinh softening parameter. ``Q`` must be greater than 0.
        Default is 8.

    Notes
    -----
    Based on the asinh stretch presented in Lupton et al. 2004
    (https://ui.adsabs.harvard.edu/abs/2004PASP..116..133L).

    Examples
    --------
    .. plot::
        :show-source-link:

        import numpy as np
        from astropy.visualization import LuptonAsinhStretch
        from matplotlib import pyplot as plt

        fig, ax = plt.subplots(nrows=2, ncols=2, figsize=(8, 8),
                               layout='constrained')
        ax = ax.ravel()

        x = np.linspace(0, 1, 100)
        stretches = (0.05, 0.1, 0.2, 0.5, 1, 5, 10)

        Qs = (2, 5, 8, 10)
        for i, Q in enumerate(Qs):
            for st in stretches:
                stretch = LuptonAsinhStretch(stretch=st, Q=Q)
                label = f'{st=}'
                ax[i].plot(x, stretch(x, clip=True), label=label)

            ax[i].axis('equal')
            ax[i].plot(x, x, ls='dotted', color='k', alpha=0.3)
            ax[i].set_xlim(0, 1)
            ax[i].set_ylim(0, 1)
            ax[i].set_xlabel('Input Value')
            ax[i].set_ylabel('Output Value')
            ax[i].set_title(f'{stretch.__class__.__name__}, {Q=}')
            ax[i].legend(loc='lower right', fontsize=8)
    """

    def __init__(self, stretch=5, Q=8):
        super().__init__()

        if stretch < 0:
            raise ValueError(f"Stretch must be non-negative! {stretch=}")
        if Q < 0:
            raise ValueError(f"Q must be non-negative! {Q=}")

        # 32bit floating point machine epsilon; sys.float_info.epsilon is 64bit
        epsilon = 1.0 / 2**23
        if abs(Q) < epsilon:
            Q = 0.1
        else:
            Qmax = 1e10
            if Q > Qmax:
                Q = Qmax

        self.stretch = stretch
        self.Q = Q

        frac = 0.1
        self._slope = frac / np.arcsinh(frac * Q)
        self._soften = Q / float(stretch)

    def __call__(self, values, clip=False, out=None):
        values = _stretch_prepare(values, clip=clip, out=out)
        np.multiply(values, self._soften, out=values)
        np.arcsinh(values, out=values)
        np.multiply(values, self._slope, out=values)
        return values


class LuptonAsinhZscaleStretch(LuptonAsinhStretch):
    r"""
    A modified asinh stretch, where the linear stretch is calculated using
    zscale.

    The stretch is given by:

    .. math::
        & y = {\rm asinh}\left(\frac{Q * x}{stretch}\right) *
        \frac{frac}{{\rm asinh}(frac * Q)} \\
        & frac = 0.1 \\
        & stretch = z2 - z1

    Parameters
    ----------
    image1 : ndarray or array-like
        The image to analyse, or a list of 3 images to be converted to
        an intensity image.

    Q : float, optional
        The asinh softening parameter. ``Q`` must be greater than 0.
        Default is 8.

    pedestal : or array-like, optional
        The value, or array of 3 values, to subtract from the images(s)
        before determining the zscaling. Default is None (nothing subtracted).

    """

    def __init__(self, image, Q=8, pedestal=None):
        # copy because of in-place operations after
        image = np.array(image, copy=True, dtype=float)

        _raiseerr = False
        if len(image.shape) == 2:
            image = [image]
        elif len(image.shape) == 3:
            if image.shape[0] != 3:
                _raiseerr = True
        else:
            _raiseerr = True

        if _raiseerr:
            raise ValueError(
                "Input 'image' must be a single image "
                "or a stack/3xMxN array of 3 images! "
                f"{image.shape=}"
            )

        image = list(image)  # needs to be mutable

        if pedestal is not None:
            try:
                len(pedestal)
            except TypeError:
                pedestal = 3 * [pedestal]

            if len(pedestal) != 3:
                raise ValueError(
                    "pedestal must be 1 or 3 values, matching the image input."
                )
            for i, im in enumerate(image):
                if pedestal[i] != 0.0:
                    image[i] = im - pedestal[i]  # n.b. a copy

        image = compute_intensity(*image)
        zscale_limits = ZScaleInterval().get_limits(image)

        _stretch = zscale_limits[1] - zscale_limits[0]

        self._image = image

        super().__init__(stretch=_stretch, Q=Q)


class RGBImageMappingLupton(RGBImageMapping):
    """
    Class to map red, blue, green images into either a normalized float or
    an 8-bit image, by performing optional clipping and applying
    a scaling function to each band in non-independent manner that depends
    on the other bands, following the scaling scheme presented in
    Lupton et al. 2004.

    Parameters
    ----------
    interval : `~astropy.visualization.BaseInterval` subclass instance or array-like, optional
        The interval object to apply to the data (either a single instance or
        an array for R, G, B). Default is
        `~astropy.visualization.ManualInterval`.
    stretch : `~astropy.visualization.BaseStretch` subclass instance
        The stretch object to apply to the data. The default is
        `~astropy.visualization.AsinhLuptonStretch`.

    """

    def __init__(
        self,
        interval=ManualInterval(vmin=0, vmax=None),
        stretch=LuptonAsinhStretch(stretch=5, Q=8),
    ):
        super().__init__(interval=interval, stretch=stretch)
        self._pixmax = 1.0

    def intensity(self, image_r, image_g, image_b):
        """
        Return the total intensity from the red, blue, and green intensities.
        This is a naive computation, and may be overridden by subclasses.

        Parameters
        ----------
        image_r : ndarray
            Intensity of image to be mapped to red; or total intensity if
            ``image_g`` and ``image_b`` are None.
        image_g : ndarray, optional
            Intensity of image to be mapped to green.
        image_b : ndarray, optional
            Intensity of image to be mapped to blue.

        Returns
        -------
        intensity : ndarray
            Total intensity from the red, blue and green intensities, or
            ``image_r`` if green and blue images are not provided.

        """
        return compute_intensity(image_r, image_g, image_b)

    def apply_mappings(self, image_r, image_g, image_b):
        """
        Apply mapping stretch and intervals to convert images image_r, image_g,
        and image_b to a triplet of normalized images, following the scaling
        scheme presented in Lupton et al. 2004.

        Compared to astropy's ImageNormalize which first normalizes images
        by cropping and linearly mapping onto [0.,1.] and then applies
        a specified stretch algorithm, the Lupton et al. algorithm applies
        stretching to an multi-color intensity and then computes per-band
        scaled images with bound cropping.

        This is modified here by allowing for different minimum values
        for each of the input r, g, b images, and then computing
        the intensity on the subtracted images.

        Parameters
        ----------
        image_r : ndarray
            Intensity of image to be mapped to red
        image_g : ndarray
            Intensity of image to be mapped to green.
        image_b : ndarray
            Intensity of image to be mapped to blue.

        Returns
        -------
        image_rgb : ndarray
            Triplet of mapped images based on the specified (per-band)
            intervals and the stretch function

        Notes
        -----
        The Lupton et al 2004 algorithm is computed with the following steps:

        1. Shift each band with the minimum values
        2. Compute the intensity I and stretched intensity f(I)
        3. Compute the ratio of the stretched intensity to intensity f(I)/I,
        and clip to a lower bound of 0
        4. Compute the scaled band images by multiplying with the ratio f(I)/I
        5. Clip each band to a lower bound of 0
        6. Scale down pixels where max(R,G,B)>1 by the value max(R,G,B)

        """
        image_r = np.array(image_r, copy=True)
        image_g = np.array(image_g, copy=True)
        image_b = np.array(image_b, copy=True)

        # Subtract per-band minima
        image_rgb = [image_r, image_g, image_b]
        for i, img in enumerate(image_rgb):
            vmin, _ = self.intervals[i].get_limits(img)
            image_rgb[i] = np.subtract(img, vmin)

        image_rgb = np.asarray(image_rgb)

        # Determine the intensity and streteched intensity
        Int = self.intensity(*image_rgb)
        fI = self.stretch(Int, clip=False)

        # Get normalized fI, and clip to lower bound of 0:
        fInorm = np.where(Int <= 0, 0, np.true_divide(fI, Int))

        # Compute X = x * f(I) / I for each filter x=(r,g,b)
        np.multiply(image_rgb, fInorm, out=image_rgb)

        # Clip individual bands to minimum of 0, as
        # individual bands can be < 0 even if fI/I isn't.
        image_rgb = np.clip(image_rgb, 0.0, None)

        # Determine the max of all 3 bands at each position
        maxRGB = np.max(image_rgb, axis=0)

        with np.errstate(invalid="ignore", divide="ignore"):
            image_rgb = np.where(
                maxRGB > self._pixmax,
                np.true_divide(image_rgb * self._pixmax, maxRGB),
                image_rgb,
            )

        return np.asarray(image_rgb)


def make_lupton_rgb(
    image_r,
    image_g,
    image_b,
    interval=None,
    stretch_object=None,
    minimum=None,
    stretch=5,
    Q=8,
    filename=None,
    output_dtype=np.uint8,
):
    r"""
    Return a Red/Green/Blue color image from 3 images using interconnected
    band scaling, and an arbitrary stretch function (by default, an asinh stretch).
    The input images can be int or float, and in any range or bit-depth.

    For a more detailed look at the use of this method, see the document
    :ref:`astropy:astropy-visualization-rgb`.

    Parameters
    ----------
    image_r : ndarray
        Image to map to red.
    image_g : ndarray
        Image to map to green.
    image_b : ndarray
        Image to map to blue.
    interval : `~astropy.visualization.BaseInterval` subclass instance or array-like, optional
        The interval object to apply to the data (either a single instance or
        an array for R, G, B). Default is
        `~astropy.visualization.ManualInterval` with vmin=0.
    stretch_object : `~astropy.visualization.BaseStretch` subclass instance, optional
        The stretch object to apply to the data. If set, the input values of
        ``minimum``, ``stretch``, and ``Q`` will be ignored.
        For the Lupton scheme, this would be an instance of
        `~astropy.visualization.LuptonAsinhStretch`, but alternatively
        `~astropy.visualization.LuptonAsinhZscaleStretch` or some other
        stretch can be used.
    minimum : float or array-like, optional
        Deprecated. Intensity that should be mapped to black (a scalar or
        array of R, G, B). If `None`, each image's minimum value is used.
        Default is None.
    stretch : float, optional
        The linear stretch of the image. Default is 5
    Q : float, optional
        The asinh softening parameter. Default is 8.
    filename : str, optional
        Write the resulting RGB image to a file (file type determined
        from extension).
    output_dtype : numpy scalar type, optional
        Image output data type. Default is np.uint8.

    Returns
    -------
    rgb : ndarray
        RGB color image as an NxNx3 numpy array, with the specified
        data type format

    """
    if stretch_object is None:
        stretch_object = LuptonAsinhStretch(stretch=stretch, Q=Q)

    if interval is None:
        # Only use minimum if interval is not specified:
        if minimum is not None:
            # Backwards compatibility:
            try:
                len(minimum)
            except TypeError:
                minimum = 3 * [minimum]
            if len(minimum) != 3:
                raise ValueError("please provide 1 or 3 values for minimum.")
            interval = []
            for i in range(3):
                interval.append(ManualInterval(vmin=minimum[i], vmax=None))
        else:
            # Default option:
            interval = ManualInterval(vmin=0, vmax=None)

    lup_map = RGBImageMappingLupton(
        interval=interval,
        stretch=stretch_object,
    )
    rgb = lup_map.make_rgb_image(image_r, image_g, image_b, output_dtype=output_dtype)

    if filename:
        import matplotlib.image

        matplotlib.image.imsave(filename, rgb, origin="lower")

    return rgb



import numpy as np
import pytest

from astropy.io import ascii
from astropy.io.ascii.qdp import _get_lines_from_file, _read_table_qdp, _write_table_qdp
from astropy.table import Column, MaskedColumn, Table
from astropy.utils.exceptions import AstropyUserWarning


def test_get_tables_from_qdp_file(tmp_path):
    example_qdp = """
    ! Swift/XRT hardness ratio of trigger: XXXX, name: BUBU X-2
    ! Columns are as labelled
    READ TERR 1
    READ SERR 2
    ! WT -- hard data
    !MJD            Err (pos)       Err(neg)        Rate            Error
    53000.123456 2.37847222222222e-05    -2.37847222222222e-05   -0.212439       0.212439
    55045.099887 1.14467592592593e-05    -1.14467592592593e-05   0.000000        0.000000
    NO NO NO NO NO
    ! WT -- soft data
    !MJD            Err (pos)       Err(neg)        Rate            Error
    53000.123456 2.37847222222222e-05    -2.37847222222222e-05   0.726155        0.583890
    55045.099887 1.14467592592593e-05    -1.14467592592593e-05   2.410935        1.393592
    NO NO NO NO NO
    ! WT -- hardness ratio
    !MJD            Err (pos)       Err(neg)        Rate            Error
    53000.123456 2.37847222222222e-05    -2.37847222222222e-05   -0.292553       -0.374935
    55045.099887 1.14467592592593e-05    -1.14467592592593e-05   0.000000        -nan
    """

    path = tmp_path / "test.qdp"

    with open(path, "w") as fp:
        print(example_qdp, file=fp)

    table0 = _read_table_qdp(fp.name, names=["MJD", "Rate"], table_id=0)
    assert table0.meta["initial_comments"][0].startswith("Swift")
    assert table0.meta["comments"][0].startswith("WT -- hard data")
    table2 = _read_table_qdp(fp.name, names=["MJD", "Rate"], table_id=2)
    assert table2.meta["initial_comments"][0].startswith("Swift")
    assert table2.meta["comments"][0].startswith("WT -- hardness")
    assert np.isclose(table2["MJD_nerr"][0], -2.37847222222222e-05)


def lowercase_header(value):
    """Make every non-comment line lower case."""
    lines = []
    for line in value.splitlines():
        if not line.startswith("!"):
            line = line.lower()
        lines.append(line)
    return "\n".join(lines)


@pytest.mark.parametrize("lowercase", [False, True])
def test_roundtrip(tmp_path, lowercase):
    example_qdp = """
    ! Swift/XRT hardness ratio of trigger: XXXX, name: BUBU X-2
    ! Columns are as labelled
    READ TERR 1
    READ SERR 2
    ! WT -- hard data
    !MJD            Err (pos)       Err(neg)        Rate            Error
    53000.123456 2.37847222222222e-05    -2.37847222222222e-05   NO       0.212439
    55045.099887 1.14467592592593e-05    -1.14467592592593e-05   0.000000        0.000000
    NO NO NO NO NO
    ! WT -- soft data
    !MJD            Err (pos)       Err(neg)        Rate            Error
    53000.123456 2.37847222222222e-05    -2.37847222222222e-05   0.726155        0.583890
    55045.099887 1.14467592592593e-05    -1.14467592592593e-05   2.410935        1.393592
    NO NO NO NO NO
    ! WT -- hardness ratio
    !MJD            Err (pos)       Err(neg)        Rate            Error
    53000.123456 2.37847222222222e-05    -2.37847222222222e-05   -0.292553       -0.374935
    55045.099887 1.14467592592593e-05    -1.14467592592593e-05   0.000000        NO
    ! Add command, just to raise the warning.
    READ TERR 1
    ! WT -- whatever
    !MJD            Err (pos)       Err(neg)        Rate            Error
    53000.123456 2.37847222222222e-05    -2.37847222222222e-05   -0.292553       -0.374935
    NO 1.14467592592593e-05    -1.14467592592593e-05   0.000000        NO
    """
    if lowercase:
        example_qdp = lowercase_header(example_qdp)

    path = str(tmp_path / "test.qdp")
    path2 = str(tmp_path / "test2.qdp")

    with open(path, "w") as fp:
        print(example_qdp, file=fp)
    with pytest.warns(AstropyUserWarning) as record:
        table = _read_table_qdp(path, names=["MJD", "Rate"], table_id=0)
    assert np.any(
        [
            "This file contains multiple command blocks" in r.message.args[0]
            for r in record
        ]
    )

    _write_table_qdp(table, path2)

    new_table = _read_table_qdp(path2, names=["MJD", "Rate"], table_id=0)

    for col in new_table.colnames:
        is_masked = np.array([np.ma.is_masked(val) for val in new_table[col]])
        if np.any(is_masked):
            # All NaN values are read as such.
            assert np.ma.is_masked(table[col][is_masked])

        is_nan = np.array(
            [(not np.ma.is_masked(val) and np.isnan(val)) for val in new_table[col]]
        )
        # All non-NaN values are the same
        assert np.allclose(new_table[col][~is_nan], table[col][~is_nan])
        if np.any(is_nan):
            # All NaN values are read as such.
            assert np.isnan(table[col][is_nan])
    assert np.allclose(new_table["MJD_perr"], [2.378472e-05, 1.1446759e-05])

    for meta_name in ["initial_comments", "comments"]:
        assert meta_name in new_table.meta


def test_read_example():
    example_qdp = """
        ! Initial comment line 1
        ! Initial comment line 2
        READ TERR 1
        READ SERR 3
        ! Table 0 comment
        !a a(pos) a(neg) b c ce d
        53000.5   0.25  -0.5   1  1.5  3.5 2
        54000.5   1.25  -1.5   2  2.5  4.5 3
        NO NO NO NO NO
        ! Table 1 comment
        !a a(pos) a(neg) b c ce d
        54000.5   2.25  -2.5   NO  3.5  5.5 5
        55000.5   3.25  -3.5   4  4.5  6.5 nan
        """
    dat = ascii.read(example_qdp, format="qdp", table_id=1, names=["a", "b", "c", "d"])
    t = Table.read(
        example_qdp, format="ascii.qdp", table_id=1, names=["a", "b", "c", "d"]
    )

    assert np.allclose(t["a"], [54000, 55000])
    assert t["c_err"][0] == 5.5
    assert np.ma.is_masked(t["b"][0])
    assert np.isnan(t["d"][1])

    for col1, col2 in zip(t.itercols(), dat.itercols()):
        assert np.allclose(col1, col2, equal_nan=True)


def test_roundtrip_example(tmp_path):
    example_qdp = """
        ! Initial comment line 1
        ! Initial comment line 2
        READ TERR 1
        READ SERR 3
        ! Table 0 comment
        !a a(pos) a(neg) b c ce d
        53000.5   0.25  -0.5   1  1.5  3.5 2
        54000.5   1.25  -1.5   2  2.5  4.5 3
        NO NO NO NO NO
        ! Table 1 comment
        !a a(pos) a(neg) b c ce d
        54000.5   2.25  -2.5   NO  3.5  5.5 5
        55000.5   3.25  -3.5   4  4.5  6.5 nan
        """
    test_file = tmp_path / "test.qdp"

    t = Table.read(
        example_qdp, format="ascii.qdp", table_id=1, names=["a", "b", "c", "d"]
    )
    t.write(test_file, err_specs={"terr": [1], "serr": [3]})
    t2 = Table.read(test_file, names=["a", "b", "c", "d"], table_id=0)

    for col1, col2 in zip(t.itercols(), t2.itercols()):
        assert np.allclose(col1, col2, equal_nan=True)


def test_roundtrip_example_comma(tmp_path):
    example_qdp = """
        ! Initial comment line 1
        ! Initial comment line 2
        READ TERR 1
        READ SERR 3
        ! Table 0 comment
        !a,a(pos),a(neg),b,c,ce,d
        53000.5,0.25,-0.5,1,1.5,3.5,2
        54000.5,1.25,-1.5,2,2.5,4.5,3
        NO,NO,NO,NO,NO
        ! Table 1 comment
        !a,a(pos),a(neg),b,c,ce,d
        54000.5,2.25,-2.5,NO,3.5,5.5,5
        55000.5,3.25,-3.5,4,4.5,6.5,nan
        """
    test_file = tmp_path / "test.qdp"

    t = Table.read(
        example_qdp, format="ascii.qdp", table_id=1, names=["a", "b", "c", "d"], sep=","
    )
    t.write(test_file, err_specs={"terr": [1], "serr": [3]})
    t2 = Table.read(test_file, names=["a", "b", "c", "d"], table_id=0)

    # t.values_equal(t2)
    for col1, col2 in zip(t.itercols(), t2.itercols()):
        assert np.allclose(col1, col2, equal_nan=True)


def test_read_write_simple(tmp_path):
    test_file = tmp_path / "test.qdp"
    t1 = Table()
    t1.add_column(Column(name="a", data=[1, 2, 3, 4]))
    t1.add_column(
        MaskedColumn(
            data=[4.0, np.nan, 3.0, 1.0], name="b", mask=[False, False, False, True]
        )
    )
    t1.write(test_file, format="ascii.qdp")
    with pytest.warns(
        AstropyUserWarning,
        match=r"^table_id not specified\. Reading the first available table$",
    ) as record:
        t2 = Table.read(test_file, format="ascii.qdp")
    assert np.any(
        [
            "table_id not specified. Reading the first available table"
            in r.message.args[0]
            for r in record
        ]
    )

    assert np.allclose(t2["col1"], t1["a"])
    assert np.all(t2["col1"] == t1["a"])

    good = ~np.isnan(t1["b"])
    assert np.allclose(t2["col2"][good], t1["b"][good])


def test_read_write_simple_specify_name(tmp_path):
    test_file = tmp_path / "test.qdp"
    t1 = Table()
    t1.add_column(Column(name="a", data=[1, 2, 3]))
    # Give a non-None err_specs
    t1.write(test_file, format="ascii.qdp")
    t2 = Table.read(test_file, table_id=0, format="ascii.qdp", names=["a"])
    assert np.all(t2["a"] == t1["a"])


def test_get_lines_from_qdp(tmp_path):
    test_file = str(tmp_path / "test.qdp")
    text_string = "A\nB"
    text_output = _get_lines_from_file(text_string)
    with open(test_file, "w") as fobj:
        print(text_string, file=fobj)
    file_output = _get_lines_from_file(test_file)
    list_output = _get_lines_from_file(["A", "B"])
    for i, line in enumerate(["A", "B"]):
        assert file_output[i] == line
        assert list_output[i] == line
        assert text_output[i] == line



# Licensed under a 3-clause BSD style license - see LICENSE.rst

import math

import erfa
import numpy as np

from astropy import units as u
from astropy.coordinates.attributes import TimeAttribute
from astropy.coordinates.baseframe import base_doc, frame_transform_graph
from astropy.coordinates.matrix_utilities import rotation_matrix
from astropy.coordinates.representation import (
    CartesianRepresentation,
    UnitSphericalRepresentation,
)
from astropy.coordinates.transformations import (
    DynamicMatrixTransform,
    FunctionTransformWithFiniteDifference,
)
from astropy.time import Time
from astropy.utils.compat import COPY_IF_NEEDED
from astropy.utils.decorators import format_doc

from .baseradec import BaseRADecFrame, doc_components
from .utils import EQUINOX_B1950

__all__ = ["FK4", "FK4NoETerms"]

jd1950 = Time("B1950").jd

doc_footer_fk4 = """
    Other parameters
    ----------------
    equinox : `~astropy.time.Time`
        The equinox of this frame.
    obstime : `~astropy.time.Time`
        The time this frame was observed.  If ``None``, will be the same as
        ``equinox``.
"""


@format_doc(base_doc, components=doc_components, footer=doc_footer_fk4)
class FK4(BaseRADecFrame):
    """
    A coordinate or frame in the FK4 system.

    Note that this is a barycentric version of FK4 - that is, the origin for
    this frame is the Solar System Barycenter, *not* the Earth geocenter.

    The frame attributes are listed under **Other Parameters**.
    """

    equinox = TimeAttribute(default=EQUINOX_B1950, doc="The equinox time")
    obstime = TimeAttribute(
        default=None,
        secondary_attribute="equinox",
        doc="The reference time (e.g., time of observation)",
    )


# the "self" transform


@frame_transform_graph.transform(FunctionTransformWithFiniteDifference, FK4, FK4)
def fk4_to_fk4(fk4coord1, fk4frame2):
    # deceptively complicated: need to transform to No E-terms FK4, precess, and
    # then come back, because precession is non-trivial with E-terms
    fnoe_w_eqx1 = fk4coord1.transform_to(FK4NoETerms(equinox=fk4coord1.equinox))
    fnoe_w_eqx2 = fnoe_w_eqx1.transform_to(FK4NoETerms(equinox=fk4frame2.equinox))
    return fnoe_w_eqx2.transform_to(fk4frame2)


@format_doc(base_doc, components=doc_components, footer=doc_footer_fk4)
class FK4NoETerms(BaseRADecFrame):
    """
    A coordinate or frame in the FK4 system, but with the E-terms of aberration
    removed.

    The frame attributes are listed under **Other Parameters**.
    """

    equinox = TimeAttribute(default=EQUINOX_B1950, doc="The equinox time")
    obstime = TimeAttribute(
        default=None,
        secondary_attribute="equinox",
        doc="The reference time (e.g., time of observation)",
    )

    @staticmethod
    def _precession_matrix(oldequinox, newequinox):
        """
        Compute and return the precession matrix for FK4 using Newcomb's method.
        Used inside some of the transformation functions.

        Parameters
        ----------
        oldequinox : `~astropy.time.Time`
            The equinox to precess from.
        newequinox : `~astropy.time.Time`
            The equinox to precess to.

        Returns
        -------
        newcoord : array
            The precession matrix to transform to the new equinox
        """
        # tropical years
        t1 = (oldequinox.byear - 1850.0) / 1000.0
        dt = (newequinox.byear - 1850.0) / 1000.0 - t1
        dt_over_3600 = dt / 3600

        z1 = zeta1 = (0.060 * t1 + 139.720) * t1 + 23035.545
        zeta2 = -0.27 * t1 + 30.240
        zeta = ((17.995 * dt + zeta2) * dt + zeta1) * dt_over_3600

        z2 = 109.480 + 0.39 * t1
        z = ((18.325 * dt + z2) * dt + z1) * dt_over_3600

        theta1 = (-0.37 * t1 - 85.29) * t1 + 20051.12
        theta2 = -0.37 * t1 - 42.65
        theta = ((-41.8 * dt + theta2) * dt + theta1) * dt_over_3600

        return (
            rotation_matrix(-z, "z")
            @ rotation_matrix(theta, "y")
            @ rotation_matrix(-zeta, "z")
        )


# the "self" transform


@frame_transform_graph.transform(DynamicMatrixTransform, FK4NoETerms, FK4NoETerms)
def fk4noe_to_fk4noe(fk4necoord1, fk4neframe2):
    return fk4necoord1._precession_matrix(fk4necoord1.equinox, fk4neframe2.equinox)


# FK4-NO-E to/from FK4 ----------------------------->
# Unlike other frames, this module include *two* frame classes for FK4
# coordinates - one including the E-terms of aberration (FK4), and
# one not including them (FK4NoETerms). The following functions
# implement the transformation between these two.
def fk4_e_terms(equinox):
    """
    Return the e-terms of aberration vector.

    Parameters
    ----------
    equinox : Time object
        The equinox for which to compute the e-terms
    """
    # Constant of aberration at J2000; from Explanatory Supplement to the
    # Astronomical Almanac (Seidelmann, 2005).
    k = 0.0056932  # in degrees (v_earth/c ~ 1e-4 rad ~ 0.0057 deg)

    # Explanatory Supplement to the Astronomical Almanac: P. Kenneth
    #  Seidelmann (ed), University Science Books (1992).
    T = (equinox.jd - jd1950) / 36525.0
    # Eccentricity of the Earth's orbit
    ek = math.radians(k) * ((-0.000000126 * T - 0.00004193) * T + 0.01673011)
    # Mean longitude of perigee of the solar orbit
    g = np.radians((((0.012 * T + 1.65) * T + 6190.67) * T + 1015489.951) / 3600.0)
    minus_ek_cos_g = -ek * np.cos(g)
    # Obliquity of the ecliptic
    o = erfa.obl80(equinox.jd, 0)

    return (ek * np.sin(g), minus_ek_cos_g * np.cos(o), minus_ek_cos_g * np.sin(o))


@frame_transform_graph.transform(
    FunctionTransformWithFiniteDifference, FK4, FK4NoETerms
)
def fk4_to_fk4_no_e(fk4coord, fk4noeframe):
    # Extract cartesian vector
    rep = fk4coord.cartesian

    # Find distance (for re-normalization)
    d_orig = rep.norm()
    rep /= d_orig

    # Apply E-terms of aberration. Note that this depends on the equinox (not
    # the observing time/epoch) of the coordinates. See issue #1496 for a
    # discussion of this.
    eterms_a = CartesianRepresentation(
        u.Quantity(
            fk4_e_terms(fk4coord.equinox),
            u.dimensionless_unscaled,
            copy=COPY_IF_NEEDED,
        ),
        copy=False,
    )
    rep = rep - eterms_a + eterms_a.dot(rep) * rep

    # Find new distance (for re-normalization)
    d_new = rep.norm()

    # Renormalize
    rep *= d_orig / d_new

    # now re-cast into an appropriate Representation, and precess if need be
    if isinstance(fk4coord.data, UnitSphericalRepresentation):
        rep = rep.represent_as(UnitSphericalRepresentation)

    # if no obstime was given in the new frame, use the old one for consistency
    newobstime = (
        fk4coord._obstime if fk4noeframe._obstime is None else fk4noeframe._obstime
    )

    fk4noe = FK4NoETerms(rep, equinox=fk4coord.equinox, obstime=newobstime)
    if fk4coord.equinox != fk4noeframe.equinox:
        # precession
        fk4noe = fk4noe.transform_to(fk4noeframe)
    return fk4noe


@frame_transform_graph.transform(
    FunctionTransformWithFiniteDifference, FK4NoETerms, FK4
)
def fk4_no_e_to_fk4(fk4noecoord, fk4frame):
    # first precess, if necessary
    if fk4noecoord.equinox != fk4frame.equinox:
        fk4noe_w_fk4equinox = FK4NoETerms(
            equinox=fk4frame.equinox, obstime=fk4noecoord.obstime
        )
        fk4noecoord = fk4noecoord.transform_to(fk4noe_w_fk4equinox)

    # Extract cartesian vector
    rep = fk4noecoord.cartesian

    # Find distance (for re-normalization)
    d_orig = rep.norm()
    rep /= d_orig

    # Apply E-terms of aberration. Note that this depends on the equinox (not
    # the observing time/epoch) of the coordinates. See issue #1496 for a
    # discussion of this.
    eterms_a = CartesianRepresentation(
        u.Quantity(
            fk4_e_terms(fk4noecoord.equinox),
            u.dimensionless_unscaled,
            copy=COPY_IF_NEEDED,
        ),
        copy=False,
    )

    eterms_a_plus_rep_ini = eterms_a + rep
    for _ in range(10):
        rep = eterms_a_plus_rep_ini / (1.0 + eterms_a.dot(rep))

    # Find new distance (for re-normalization)
    d_new = rep.norm()

    # Renormalize
    rep *= d_orig / d_new

    # now re-cast into an appropriate Representation, and precess if need be
    if isinstance(fk4noecoord.data, UnitSphericalRepresentation):
        rep = rep.represent_as(UnitSphericalRepresentation)

    return fk4frame.realize_frame(rep)



# Licensed under a 3-clause BSD style license - see LICENSE.rst

import pytest

from astropy.cosmology._src.flrw.base import quad
from astropy.utils.compat.optional_deps import HAS_SCIPY


@pytest.mark.skipif(HAS_SCIPY, reason="scipy is installed")
def test_optional_deps_functions():
    """Test stand-in functions when optional dependencies not installed."""
    with pytest.raises(ModuleNotFoundError, match="No module named 'scipy.integrate'"):
        quad()



# Licensed under a 3-clause BSD style license - see LICENSE.rst

import pkgutil
import subprocess
import sys
from pathlib import Path
from textwrap import dedent
from types import ModuleType

import pytest

import astropy


@pytest.mark.parametrize(
    "subpkg",
    [subpkg.name for subpkg in pkgutil.walk_packages(astropy.__path__) if subpkg.ispkg],
)
def test_imports(subpkg):
    """
    This just imports all top-level sub-packages in astropy, making sure they don't have any
    dependencies that sneak through
    """
    subprocess.run([sys.executable, "-c", f"from astropy import {subpkg}"], check=True)


def test_toplevel_namespace():
    import astropy

    d = dir(astropy)
    assert "os" not in d
    assert "log" in d
    assert "test" in d
    assert "sys" not in d


def test_toplevel_lazy_imports():
    # Check that subpackages are loaded on demand.
    cmd = dedent("""
    import astropy, sys
    assert 'astropy.units' not in sys.modules
    astropy.units
    assert 'astropy.units' in sys.modules
    """)
    cp = subprocess.check_call([sys.executable, "-c", cmd])
    assert cp == 0


def test_toplevel_attribute_error():
    # Ensure that our __getattr__ does not leak an import error or so.
    with pytest.raises(AttributeError, match="module 'astropy' has no"):
        astropy.nonsense


def test_completeness_toplevel__all__():
    # Implicitly check that all items in __all__ exist.
    all_items = {getattr(astropy, attr) for attr in astropy.__all__}
    # Verify that the list of modules in __all__ is complete.
    module_names = {
        item.__name__.partition(".")[2]
        for item in all_items
        if isinstance(item, ModuleType)
    }
    module_dirs = {
        f.name
        for f in Path(astropy.__file__).parent.glob("[a-z]*")
        if f.is_dir() and f.name != "extern"
    }
    assert module_names == module_dirs



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Cosmological units and equivalencies."""

from astropy.units.docgen import generate_unit_summary as _generate_unit_summary

__all__ = [
    # redshift equivalencies
    "dimensionless_redshift",
    "littleh",
    "redshift",
    "redshift_distance",
    "redshift_hubble",
    "redshift_temperature",
    # other equivalencies
    "with_H0",
    "with_redshift",
]


from astropy.units import add_enabled_equivalencies as _add_enabled_equivalencies

from ._src.units import littleh, redshift
from ._src.units_equivalencies import (
    dimensionless_redshift,
    redshift_distance,
    redshift_hubble,
    redshift_temperature,
    with_H0,
    with_redshift,
)

# ===================================================================
# Enable the set of default equivalencies.
# If the cosmology package is imported, this is added to the list astropy-wide.
_add_enabled_equivalencies(dimensionless_redshift())


# =============================================================================
# DOCSTRING

# This generates a docstring for this module that describes all of the
# standard units defined here.
if __doc__ is not None:
    from ._src.units import _ns

    __doc__ += "\n" + _generate_unit_summary(_ns)



"""
DataFrame conversion utilities for Astropy Tables.

This module provides utility functions for converting between Astropy Tables
and various DataFrame formats (pandas, polars, pyarrow, etc) via the narwhals
library.
"""

from __future__ import annotations

import math
import warnings
from collections.abc import Mapping
from typing import TYPE_CHECKING, Any, NewType

import numpy as np

from astropy.utils.compat.optional_deps import HAS_NARWHALS, HAS_PANDAS

from .column import Column, MaskedColumn

__all__ = ["from_df", "from_pandas", "to_df", "to_pandas"]

if TYPE_CHECKING:
    from narwhals.typing import EagerAllowed, IntoBackend
    from units.typing import UnitLike

    from .table import Table

# Sentinel value to indicate pandas-like backend validation
PandasLikeSentinel = NewType("PandasLikeSentinel", object)  # Custom type
PANDAS_LIKE: PandasLikeSentinel = PandasLikeSentinel(object())

# Fixed set of integer dtypes
INTEGER_DTYPE_KINDS = frozenset({"u", "i"})


def _numpy_to_pandas_dtype(dtype: np.dtype) -> str:
    """Convert a numpy dtype to a pandas dtype string, handling nullable integers."""
    dtype_name = dtype.name
    # Special case needed for uint -> UInt
    if dtype_name.startswith("uint"):
        return "UInt" + dtype_name.removeprefix("uint")
    return dtype.name.title()


def _encode_mixins(tbl: Table) -> Table:
    """Encode mixin columns to basic columns for DataFrame compatibility."""
    from astropy.time import TimeBase, TimeDelta

    from . import serialize
    from .column import col_copy

    time_cols = [col for col in tbl.itercols() if isinstance(col, TimeBase)]
    if time_cols:
        new_cols = []
        for col in tbl.itercols():
            new_col = col_copy(col, copy_indices=False) if col.info.indices else col
            new_cols.append(new_col)
        tbl = tbl.__class__(new_cols, copy=False)

        for col in tbl.itercols():
            col.info.indices.clear()

        for col in time_cols:
            if isinstance(col, TimeDelta):
                new_col = (col.sec * 1e9).astype("timedelta64[ns]")
                nat = np.timedelta64("NaT")
            else:
                new_col = col.datetime64.copy()
                nat = np.datetime64("NaT")
            if col.masked:
                new_col[col.mask] = nat
            tbl[col.info.name] = new_col

    # Convert the table to one with no mixins, only Column objects.
    encode_tbl = serialize.represent_mixins_as_columns(tbl)
    return encode_tbl


def _get_backend_impl(backend: str):
    """Get the narwhals backend implementation."""
    if not HAS_NARWHALS:
        raise ModuleNotFoundError(
            "The narwhals library is required for generic DataFrame conversion."
        )
    import narwhals as nw

    return nw.Implementation.from_backend(backend)


def _is_pandas_like(
    backend_impl: IntoBackend[EagerAllowed] | PandasLikeSentinel,
) -> bool:
    return backend_impl is PANDAS_LIKE or (
        not isinstance(backend_impl, str)
        and getattr(backend_impl, "is_pandas_like", lambda: False)()
    )


def _validate_columns_for_backend(
    table: Table,
    *,
    backend_impl: IntoBackend[EagerAllowed] | PandasLikeSentinel = PANDAS_LIKE,
) -> None:
    """Validate that table columns are compatible with the target backend.

    backend_impl may be PANDAS_LIKE to indicate pandas-like validation.

    Raises ValueError if there is a multidimensional column with an unsupported backend.
    """
    # Check for multidimensional columns
    badcols = [name for name, col in table.columns.items() if len(col.shape) > 1]

    if not badcols:
        return

    # Check if pandas-like or pyarrow-like
    if _is_pandas_like(backend_impl) or (
        not isinstance(backend_impl, str)
        and getattr(backend_impl, "is_pyarrow", lambda: False)()
    ):
        raise ValueError(
            f"Cannot convert a table with multidimensional columns to a "
            f"pandas-like or pyarrow DataFrame. Offending columns are: {badcols}\n"
            f"One can filter out such columns using:\n"
            f"names = [name for name in tbl.colnames if len(tbl[name].shape) <= 1]\n"
            f"tbl[names].to_pandas(...)"
        )


def _handle_index_argument(
    table: Table,
    *,
    index: bool | str | None,
    backend_impl: IntoBackend[EagerAllowed] | PandasLikeSentinel,
) -> bool | str:
    """Process the index argument for DataFrame conversion."""
    if index is False:
        return False

    # Check if pandas-like, PANDAS_LIKE is reserved for pandas itself
    # Non-pandas backends don't support indexing
    if not _is_pandas_like(backend_impl):
        if index:
            raise ValueError("Indexing is only supported for pandas-like backends.")
        return False

    has_single_pk = table.primary_key is not None and len(table.primary_key) == 1

    if index in (None, True):
        if has_single_pk:
            return table.primary_key[0]
        elif index is None:
            return False
        else:
            assert index is True
            raise ValueError("index=True requires a single-column primary key.")

    elif isinstance(index, str):
        if index not in table.colnames:
            raise ValueError(f"{index!r} is not in the table columns.")
        return index

    else:
        raise TypeError(
            "index must be None, False, True, or a valid column name. "
            f"Type provided: {type(index).__name__}"
        )


def to_df(
    table: Table,
    *,
    backend: str,
    index: bool | str | None = None,
    use_nullable_int: bool = True,
):
    """Convert an Astropy Table to a DataFrame using the specified backend."""
    if not HAS_NARWHALS:
        raise ModuleNotFoundError(
            "The narwhals library is required for the generic to_df method. "
            "If you want to only convert to pandas, use the `to_pandas` method instead."
        )
    import narwhals as nw

    backend_impl = _get_backend_impl(backend)

    # Using private API while narwhals-dev/narwhals#3150 is not merged
    if not nw._utils.is_eager_allowed(backend_impl):
        raise ValueError("Must export to eager compatible DataFrame")

    # Handle index argument
    index = _handle_index_argument(table, index=index, backend_impl=backend_impl)

    # Encode mixins and validate columns
    tbl = _encode_mixins(table)
    _validate_columns_for_backend(tbl, backend_impl=backend_impl)

    # Convert to narwhals DataFrame
    array = tbl.as_array()
    array_dict = {
        n: array[n].data if tbl.has_masked_columns else array[n] for n in tbl.colnames
    }
    df_nw = nw.from_dict(array_dict, backend=backend_impl)

    # Handle masked columns
    if tbl.has_masked_columns:
        masked_cols = [
            name
            for name, col in tbl.columns.items()
            if isinstance(col, MaskedColumn) and col.mask.any()
        ]
        old_dtypes = {n: df_nw[n].dtype for n in masked_cols}

        df_nw = df_nw.with_columns(
            (nw.when(~nw.new_series(n, array[n].mask, backend=backend_impl)).then(n))
            for n in masked_cols
        )

        for n, old_dtype in old_dtypes.items():
            if old_dtype.is_integer() and not use_nullable_int:
                raise ValueError(
                    "Cannot convert masked integer columns to DataFrame without using nullable integers. "
                    f"Set use_nullable_int=True or remove the offending column: {n}."
                )
            elif old_dtype.is_float():
                df_nw = df_nw.with_columns(nw.col(n).cast(old_dtype).alias(n))

    # Convert to dataframe
    df_native = df_nw.to_native()

    # Fix pandas-like nullable integers
    if backend_impl.is_pandas_like() and tbl.has_masked_columns and use_nullable_int:
        for name in masked_cols:
            dtype = array[name].dtype
            if (dtype := array[name].dtype).kind not in INTEGER_DTYPE_KINDS:
                continue

            df_native[name] = df_native[name].astype(_numpy_to_pandas_dtype(dtype))

    # Pandas-like index
    if index:
        df_native.set_index(index, inplace=True, drop=True)

    return df_native


def from_df(
    cls: type[Table],
    df: Any,
    *,
    index: bool = False,
    units: Mapping[str, UnitLike] | None = None,
) -> Table:
    """Create an instance of ``cls`` from any narwhals-compatible DataFrame."""
    if not HAS_NARWHALS:
        raise ModuleNotFoundError(
            "The narwhals library is required for the generic from_df method. "
            "If you want to only convert from pandas, use the `from_pandas` method instead."
        )
    import narwhals as nw

    # Create output
    out = {}

    # Handle pandas index
    if index:
        if not hasattr(df, "index"):
            raise ValueError(
                "The input dataframe does not have an index. "
                "Are you trying to convert a non-pandas dataframe? "
                "Set `index=False` (or leave it unspecified) to avoid this error."
            )
        index_name = str(df.index.name or "index")
        while index_name in df.columns:
            index_name = "_" + index_name + "_"
        df.reset_index(index_name, inplace=True, drop=False)

    # Narwhals layer, must convert to eager
    df_nw = nw.from_native(df)
    if isinstance(df_nw, nw.LazyFrame):
        df_nw = df_nw.collect()

    # Handle units
    if units is None:
        units = {}
    elif not isinstance(units, Mapping):
        raise TypeError(
            f"Expected a Mapping from column-names to units. Got {units!r} with type {type(units)}"
        )
    not_found = set(units.keys()) - set(df_nw.columns)
    if not_found:
        warnings.warn(f"`units` contains additional columns: {not_found}")

    # Iterate over Narwhals columns
    for column in df_nw.iter_columns():
        # Unpack relevant data
        name = column.name
        dtype = column.dtype
        nw_mask = column.is_null()
        unit = units.get(name)

        # Check first if the dtype is supported
        if isinstance(dtype, nw.Int128):
            raise ValueError(
                "Astropy Table does not support narwhals.Int128, please use a smaller integer type."
            )
        data = column.to_numpy()

        # Handle nullable integers
        if dtype.is_integer() and nw_mask.any():
            nullfilled = column.fill_null(0).to_numpy()
            out[name] = MaskedColumn(
                data=nullfilled, mask=nw_mask.to_numpy(), unit=unit, copy=False
            )
            continue

        # Handle datetime columns
        elif isinstance(dtype, nw.Datetime):
            from astropy.time import Time

            datetime = Time(data, format="datetime64")
            datetime.format = "isot"
            out[name] = datetime
            continue

        # Handle timedelta columns
        elif isinstance(dtype, nw.Duration):
            from astropy.time import TimeDelta

            duration = (
                data.astype("timedelta64[ns]").astype(np.float64, copy=False) / 1e9
            )
            out[name] = TimeDelta(duration, format="sec")
            if nw_mask.any():
                out[name][nw_mask.to_numpy()] = np.ma.masked
            continue

        # Handle string-like columns
        elif isinstance(dtype, (nw.String, nw.Binary, nw.Object)):
            if data.dtype.kind == "O":
                ts = (str, bytes)
                # If all elements of an object array are string-like or None or np.nan
                # then coerce back to a native numpy str/unicode array.
                if all((x is None) or isinstance(x, ts) or math.isnan(x) for x in data):
                    # Force any missing (null) values to b''.  Numpy will
                    # upcast to str/unicode as needed. We go via a list to
                    # avoid replacing objects in a view of the pandas array and
                    # to ensure numpy initializes to string or bytes correctly.
                    data = np.array(
                        [b"" if m else d for (d, m) in zip(data, nw_mask.to_numpy())]
                    )

        if nw_mask.any():
            out[name] = MaskedColumn(
                data=data, mask=nw_mask.to_numpy(), unit=unit, copy=False
            )
        else:
            out[name] = Column(data=data, unit=unit, copy=False)

    return cls(out)


def to_pandas(
    table: Table, *, index: bool | str | None = None, use_nullable_int: bool = True
):
    """Convert an Astropy Table to a pandas DataFrame.

    This mirrors the previous DataFrameConverter.to_pandas method but as a
    module-level function.
    """
    if not HAS_PANDAS:
        raise ModuleNotFoundError(
            "pandas is required for to_pandas conversion. "
            "Install with: pip install pandas"
        )
    from pandas import DataFrame, Series

    # Handle index argument (pandas-specific logic)
    index = _handle_index_argument(
        table, index=index, backend_impl=PANDAS_LIKE
    )  # PANDAS_LIKE for pandas validation

    # Encode mixins and validate columns
    tbl = _encode_mixins(table)
    _validate_columns_for_backend(tbl, backend_impl=PANDAS_LIKE)  # pandas validation

    out = {}

    for name, column in tbl.columns.items():
        if getattr(column.dtype, "isnative", True):
            out[name] = column
        else:
            out[name] = column.data.byteswap().view(column.dtype.newbyteorder("="))

        if isinstance(column, MaskedColumn) and np.any(column.mask):
            if column.dtype.kind in ["i", "u"]:
                pd_dtype = column.dtype.name
                if use_nullable_int:
                    # Convert int64 to Int64, uint32 to UInt32, etc for nullable types
                    pd_dtype = pd_dtype.replace("i", "I").replace("u", "U")
                else:
                    from pandas.errors import IntCastingNaNError

                    raise IntCastingNaNError(
                        "Cannot convert masked integer columns to DataFrame without using nullable integers. "
                        f"Set use_nullable_int=True or remove the offending column: {name}."
                    )
                out[name] = Series(out[name], dtype=pd_dtype)

            elif column.dtype.kind not in ["f", "c"]:
                out[name] = column.astype(object).filled(np.nan)

    kwargs = {}

    if index:
        idx = out.pop(index)
        kwargs["index"] = idx

        # We add the table index to Series inputs (MaskedColumn with int values) to override
        # its default RangeIndex, see #11432
        for v in out.values():
            if isinstance(v, Series):
                v.index = idx

    df = DataFrame(out, **kwargs)
    if index:
        # Explicitly set the pandas DataFrame index to the original table
        # index name.
        df.index.name = idx.info.name

    return df


def from_pandas(
    cls: type[Table],
    dataframe: Any,
    index: bool = False,
    units: Mapping[str, UnitLike] | None = None,
) -> Table:
    """Create an instance of ``cls`` from a pandas DataFrame."""
    out = {}

    names = list(dataframe.columns)
    columns = [dataframe[name] for name in names]
    datas = [np.array(column) for column in columns]
    masks = [np.array(column.isnull()) for column in columns]

    if index:
        index_name = dataframe.index.name or "index"
        while index_name in names:
            index_name = "_" + index_name + "_"
        names.insert(0, index_name)
        columns.insert(0, dataframe.index)
        datas.insert(0, np.array(dataframe.index))
        masks.insert(0, np.zeros(len(dataframe), dtype=bool))

    if units is None:
        units = [None] * len(names)
    else:
        if not isinstance(units, Mapping):
            raise TypeError(
                f"Expected a Mapping from column-names to units. Got {units!r} with type {type(units)}"
            )

        not_found = set(units.keys()) - set(names)
        if not_found:
            warnings.warn(f"`units` contains additional columns: {not_found}")

        units = [units.get(name) for name in names]

    for name, column, data, mask, unit in zip(names, columns, datas, masks, units):
        if column.dtype.kind in ["u", "i", "b"] and np.any(mask):
            # Special-case support for pandas nullable int and bool
            np_dtype = column.dtype.numpy_dtype
            data = np.zeros(shape=column.shape, dtype=np_dtype)
            data[~mask] = column[~mask]
            out[name] = MaskedColumn(
                data=data, name=name, mask=mask, unit=unit, copy=False
            )
            continue

        if data.dtype.kind == "O":
            # If all elements of an object array are string-like or np.nan
            # then coerce back to a native numpy str/unicode array.
            string_types = (str, bytes)
            nan = np.nan
            if all(isinstance(x, string_types) or x is nan for x in data):
                # Force any missing (null) values to b''.  Numpy will
                # upcast to str/unicode as needed. We go via a list to
                # avoid replacing objects in a view of the pandas array and
                # to ensure numpy initializes to string or bytes correctly.
                data = np.array([b"" if m else d for (d, m) in zip(data, mask)])

        # Numpy datetime64
        if data.dtype.kind == "M":
            from astropy.time import Time

            out[name] = Time(data, format="datetime64")
            if np.any(mask):
                out[name][mask] = np.ma.masked
            out[name].format = "isot"

        # Numpy timedelta64
        elif data.dtype.kind == "m":
            from astropy.time import TimeDelta

            data_sec = data.astype("timedelta64[ns]").astype(np.float64) / 1e9
            out[name] = TimeDelta(data_sec, format="sec")
            if np.any(mask):
                out[name][mask] = np.ma.masked

        else:
            if np.any(mask):
                out[name] = MaskedColumn(data=data, name=name, mask=mask, unit=unit)
            else:
                out[name] = Column(data=data, name=name, unit=unit)

    return cls(out)



# Licensed under a 3-clause BSD style license - see LICENSE.rst

import os
import sys
import tempfile

import numpy as np
import pytest
from numpy.testing import assert_allclose

from astropy.utils.compat.optional_deps import HAS_MATPLOTLIB, HAS_PLT
from astropy.visualization import basic_rgb
from astropy.visualization.interval import ManualInterval
from astropy.visualization.stretch import LinearStretch, LogStretch

# Set DISPLAY=True to get matplotlib imshow windows to help with debugging.
DISPLAY = False

# Override any debugging if not HAS_PLT
if HAS_PLT & DISPLAY:
    import matplotlib.pyplot as plt
elif not HAS_PLT:
    DISPLAY = False

MINSC = 0.0
MAXSC = 5.0e4
MIN = [0.0, 0.0, 0.0]
MAX = [5.0e4, 5.0e4, 4.0e4]
IX = 16
IY = 16

SCALEA = 1500.0
SHAPE = (85, 75)

# Gaussian pixel centers, peak values, and colors
_points = [[15, 15], [50, 45], [30, 30], [45, 15]]
_values = [1000, 5500, 600, 20000]
_stddevs = [1.5, 2.0, 1.0, 2.5]
_g_r = [1.0, -1.0, 1.0, 1.0]
_r_i = [2.0, -0.5, 2.5, 1.0]

_grid = np.indices(SHAPE)
IMAGER = np.zeros(SHAPE)
IMAGEG = np.zeros(SHAPE)
IMAGEB = np.zeros(SHAPE)
for p, v, std, gr, ri in zip(_points, _values, _stddevs, _g_r, _r_i):
    _gaus = np.exp(
        -(
            (_grid[0] - p[0]) ** 2 / (2.0 * std**2)
            + (_grid[1] - p[1]) ** 2 / (2.0 * std**2)
        )
    )
    IMAGER += v * np.power(10, 0.4 * ri) * _gaus
    IMAGEG += v * np.power(10, 0.4 * gr) * _gaus
    IMAGEB += v * _gaus

RSEED = 0
rng = np.random.default_rng(RSEED)
IMAGER = IMAGER + rng.normal(0, 2, SHAPE)
IMAGEG = IMAGEG + rng.normal(0, 2, SHAPE)
IMAGEB = IMAGEB + rng.normal(0, 2, SHAPE)

INCORRECT_OUTPUT_TYPES = [bool, str, np.int64, np.cdouble, "str"]


def _display_rgb(rgb, title=None):
    """Display an rgb image using matplotlib (useful for debugging)"""
    plt.imshow(rgb, interpolation="nearest", origin="lower")
    if title:
        plt.title(title)
    plt.show()
    return plt


def test_image_mapping():
    """Test creating an RGB image using a linear stretch,
    using RGBImageMapping()"""
    stretch = LinearStretch()
    interval = []
    for i in range(3):
        interval.append(ManualInterval(vmin=MIN[i], vmax=MAX[i]))
    map_ = basic_rgb.RGBImageMapping(stretch=stretch, interval=interval)
    rgb_image = map_.make_rgb_image(IMAGER, IMAGEG, IMAGEB, output_dtype=np.float64)
    for i, (min_, max_, iref_) in enumerate(
        zip(
            [0.0, 0.0, 0.0],
            [1.0, 1.0, 0.5000598388671327],
            [0.08093024185629245, 0.032216094791227695, 0.016040737174622725],
        )
    ):
        assert_allclose(rgb_image[:, :, i].min(), min_)
        assert_allclose(rgb_image[:, :, i].max(), max_)
        assert_allclose(rgb_image[IX, IY, i], iref_)
    if DISPLAY:
        _display_rgb(rgb_image, title=sys._getframe().f_code.co_name)


def test_linear():
    """Test creating an RGB image using a linear stretch,
    using individual routines"""
    interval = []
    for i in range(3):
        interval.append(ManualInterval(vmin=MIN[i], vmax=MAX[i]))
    rgb_image = basic_rgb.make_rgb(
        IMAGER,
        IMAGEG,
        IMAGEB,
        interval=interval,
        output_dtype=np.float64,
    )
    for i, (min_, max_, iref_) in enumerate(
        zip(
            [0.0, 0.0, 0.0],
            [1.0, 1.0, 0.5000598388671327],
            [0.08093024185629245, 0.032216094791227695, 0.016040737174622725],
        )
    ):
        assert_allclose(rgb_image[:, :, i].min(), min_)
        assert_allclose(rgb_image[:, :, i].max(), max_)
        assert_allclose(rgb_image[IX, IY, i], iref_)
    if DISPLAY:
        _display_rgb(rgb_image, title=sys._getframe().f_code.co_name)


def test_log():
    """Test creating an RGB image using an log stretch"""
    interval = []
    for i in range(3):
        interval.append(ManualInterval(vmin=MIN[i], vmax=MAX[i]))
    rgb_image = basic_rgb.make_rgb(
        IMAGER,
        IMAGEG,
        IMAGEB,
        interval=interval,
        stretch=LogStretch(a=SCALEA),
        output_dtype=np.float64,
    )
    for i, (min_, max_, iref_) in enumerate(
        zip(
            [0.0, 0.0, 0.0],
            [1.0, 1.0, 0.9053360156408082],
            [0.6572779418489928, 0.5330153105260111, 0.4404384627801792],
        )
    ):
        assert_allclose(rgb_image[:, :, i].min(), min_)
        assert_allclose(rgb_image[:, :, i].max(), max_)
        assert_allclose(rgb_image[IX, IY, i], iref_)
    if DISPLAY:
        _display_rgb(rgb_image, title=sys._getframe().f_code.co_name)


def test_int8():
    """Test creating an RGB image with 8-bit output format"""
    interval = []
    for i in range(3):
        interval.append(ManualInterval(vmin=MIN[i], vmax=MAX[i]))
    rgb_image = basic_rgb.make_rgb(
        IMAGER,
        IMAGEG,
        IMAGEB,
        interval=interval,
        stretch=LogStretch(a=SCALEA),
        output_dtype=np.uint8,
    )
    assert np.issubdtype(rgb_image.dtype, np.uint8)
    if DISPLAY:
        _display_rgb(rgb_image, title=sys._getframe().f_code.co_name)


def test_float64():
    """Test creating an RGB image with normalized float output format"""
    interval = []
    for i in range(3):
        interval.append(ManualInterval(vmin=MIN[i], vmax=MAX[i]))
    rgb_image = basic_rgb.make_rgb(
        IMAGER,
        IMAGEG,
        IMAGEB,
        interval=interval,
        stretch=LogStretch(a=SCALEA),
        output_dtype=np.float64,
    )
    assert np.issubdtype(rgb_image.dtype, float)
    if DISPLAY:
        _display_rgb(rgb_image, title=sys._getframe().f_code.co_name)


def test_linear_min_max():
    """Test using a min/max linear stretch determined from one image"""
    rgb_image = basic_rgb.make_rgb(
        IMAGER,
        IMAGEG,
        IMAGEB,
        interval=ManualInterval(vmin=None, vmax=None),
        output_dtype=np.float64,
    )
    for i, (min_, max_, iref_) in enumerate(
        zip(
            [0.0, 0.0, 0.0],
            [1.0, 1.0, 1.0],
            [0.08069534125307666, 0.032196043103128555, 0.032466842729915714],
        )
    ):
        assert_allclose(rgb_image[:, :, i].min(), min_)
        assert_allclose(rgb_image[:, :, i].max(), max_)
        assert_allclose(rgb_image[IX, IY, i], iref_)
    if DISPLAY:
        _display_rgb(rgb_image, title=sys._getframe().f_code.co_name)


def test_log_min_max():
    """Test using a min/max log stretch determined from one image"""
    rgb_image = basic_rgb.make_rgb(
        IMAGER,
        IMAGEG,
        IMAGEB,
        interval=ManualInterval(vmin=None, vmax=None),
        stretch=LogStretch(a=SCALEA),
        output_dtype=np.float64,
    )
    for i, (min_, max_, iref_) in enumerate(
        zip(
            [0.0, 0.0, 0.0],
            [1.0, 1.0, 1.0],
            [0.6568837677677257, 0.5329319103684619, 0.5340539629318083],
        )
    ):
        assert_allclose(rgb_image[:, :, i].min(), min_)
        assert_allclose(rgb_image[:, :, i].max(), max_)
        assert_allclose(rgb_image[IX, IY, i], iref_)
    if DISPLAY:
        _display_rgb(rgb_image, title=sys._getframe().f_code.co_name)


def test_log_scalar_interval():
    """Test creating a black+white image using a linear stretch"""
    rgb_image = basic_rgb.make_rgb(
        IMAGER,
        IMAGEG,
        IMAGEB,
        interval=ManualInterval(vmin=MINSC, vmax=MAXSC),
        stretch=LogStretch(a=SCALEA),
        output_dtype=np.float64,
    )
    for i, (min_, max_, iref_) in enumerate(
        zip(
            [0.0, 0.0, 0.0],
            [1.0, 1.0, 0.8748719461075388],
            [0.6572779418489928, 0.5330153105260111, 0.41128606174423155],
        )
    ):
        assert_allclose(rgb_image[:, :, i].min(), min_)
        assert_allclose(rgb_image[:, :, i].max(), max_)
        assert_allclose(rgb_image[IX, IY, i], iref_)
    if DISPLAY:
        _display_rgb(rgb_image, title=sys._getframe().f_code.co_name)


def test_linear_bw():
    """Test creating a black+white image using a linear stretch"""
    rgb_image = basic_rgb.make_rgb(
        IMAGER,
        IMAGER,
        IMAGER,
        interval=ManualInterval(vmin=MINSC, vmax=MAXSC),
        output_dtype=np.float64,
    )
    for i, (min_, max_, iref_) in enumerate(
        zip(
            [0.0, 0.0, 0.0],
            [1.0, 1.0, 1.0],
            [0.08093024185629245, 0.08093024185629245, 0.08093024185629245],
        )
    ):
        assert_allclose(rgb_image[:, :, i].min(), min_)
        assert_allclose(rgb_image[:, :, i].max(), max_)
        assert_allclose(rgb_image[IX, IY, i], iref_)
    if DISPLAY:
        _display_rgb(rgb_image, title=sys._getframe().f_code.co_name)


def test_log_bw():
    """Test creating a black+white image using a log stretch"""
    rgb_image = basic_rgb.make_rgb(
        IMAGER,
        IMAGER,
        IMAGER,
        interval=ManualInterval(vmin=MINSC, vmax=MAXSC),
        stretch=LogStretch(a=SCALEA),
        output_dtype=np.float64,
    )
    for i, (min_, max_, iref_) in enumerate(
        zip(
            [0.0, 0.0, 0.0],
            [1.0, 1.0, 1.0],
            [0.6572779418489928, 0.6572779418489928, 0.6572779418489928],
        )
    ):
        assert_allclose(rgb_image[:, :, i].min(), min_)
        assert_allclose(rgb_image[:, :, i].max(), max_)
        assert_allclose(rgb_image[IX, IY, i], iref_)
    if DISPLAY:
        _display_rgb(rgb_image, title=sys._getframe().f_code.co_name)


@pytest.mark.skipif(not HAS_MATPLOTLIB, reason="requires matplotlib")
def test_make_log_rgb_file():
    """Test the function that does it all"""
    interval = []
    for i in range(3):
        interval.append(ManualInterval(vmin=MIN[i], vmax=MAX[i]))
    with tempfile.NamedTemporaryFile(suffix=".png") as temp:
        red = IMAGER
        green = IMAGEG
        blue = IMAGEB
        basic_rgb.make_rgb(
            red,
            green,
            blue,
            interval=interval,
            stretch=LogStretch(a=SCALEA),
            filename=temp,
        )
        assert os.path.exists(temp.name)


@pytest.mark.skipif(not HAS_MATPLOTLIB, reason="requires matplotlib")
def test_make_linear_rgb_file():
    """Test the function that does it all"""
    interval = []
    for i in range(3):
        interval.append(ManualInterval(vmin=MIN[i], vmax=MAX[i]))
    with tempfile.NamedTemporaryFile(suffix=".png") as temp:
        red = IMAGER
        green = IMAGEG
        blue = IMAGEB
        basic_rgb.make_rgb(red, green, blue, interval=interval, filename=temp)
        assert os.path.exists(temp.name)


def test_different_shapes_asserts():
    """Test the different shape assertion"""
    with pytest.raises(ValueError, match=r"shapes must match"):
        # just swap the dimensions to get a differently-shaped 'r'
        image_r = IMAGER.reshape(SHAPE[1], SHAPE[0])
        basic_rgb.make_rgb(image_r, IMAGEG, IMAGEB, stretch=LogStretch(a=SCALEA))


def test_incorrect_interval_length():
    """Test incorrect input interval array length"""
    with pytest.raises(ValueError, match=r"3 instances for interval."):
        interval = ManualInterval(vmin=MINSC, vmax=MAXSC)
        basic_rgb.make_rgb(
            IMAGER,
            IMAGEG,
            IMAGEB,
            interval=[interval, interval],
            stretch=LogStretch(a=SCALEA),
        )


@pytest.mark.parametrize(("out_format"), INCORRECT_OUTPUT_TYPES)
def test_invalid_output_dtype(out_format):
    """Test incorrect output image format"""
    interval = []
    for i in range(3):
        interval.append(ManualInterval(vmin=MIN[i], vmax=MAX[i]))
    with pytest.raises(ValueError, match=r"'output_dtype' must be one"):
        basic_rgb.make_rgb(
            IMAGER,
            IMAGEG,
            IMAGEB,
            interval=interval,
            stretch=LogStretch(a=SCALEA),
            output_dtype=out_format,
        )



# Licensed under a 3-clause BSD style license - see LICENSE.rst

import pytest

import astropy.io.votable
import astropy.io.votable.dataorigin as dataorigin
from astropy.table import Column, Table


def __generate_votable_test():
    table = Table(
        [
            Column(name="id", data=[1, 2, 3, 4]),
            Column(name="bmag", unit="mag", data=[5.6, 7.9, 12.4, 11.3]),
        ]
    )
    return astropy.io.votable.tree.VOTableFile().from_table(table)


__TEST_PUBLISHER_NAME = "astropy"
__TEST_CREATOR1_NAME = "her"
__TEST_CREATOR2_NAME = "him"
__TEST_CONTACT = "me@mail.fr"


def __add_origin():
    vot = __generate_votable_test()
    dataorigin.add_data_origin_info(vot, "publisher", __TEST_PUBLISHER_NAME)
    dataorigin.add_data_origin_info(
        vot, "contact", __TEST_CONTACT, "Contact email address"
    )
    dataorigin.add_data_origin_info(
        vot.resources[0], "creator", __TEST_CREATOR1_NAME, "Author name"
    )
    dataorigin.add_data_origin_info(
        vot.resources[0], "creator", __TEST_CREATOR2_NAME, "Author name"
    )

    return vot


def test_dataorigin():
    vot = __add_origin()
    do = dataorigin.extract_data_origin(vot)

    assert do.query.publisher == __TEST_PUBLISHER_NAME
    assert len(do.origin) == 1 and len(do.origin[0].creator) == 2
    assert do.origin[0].creator[0] == __TEST_CREATOR1_NAME
    assert len(str(do)) > 1
    assert (do.origin[0].get_votable_element()) is not None

    dores = dataorigin.extract_data_origin(vot.resources[0])
    dot = dataorigin.extract_data_origin(vot.resources[0].tables[0])

    assert dores.origin[0].creator == do.origin[0].creator
    assert len(dot.origin) == 0

    with pytest.raises(ValueError, match="QueryOrigin contact already exists"):
        # QueryOrigin attributes are unique
        dataorigin.add_data_origin_info(vot, "contact", "othercontact")

    with pytest.raises(TypeError, match="Bad type of vot_element"):
        # contact should be global in VOFile node
        dataorigin.add_data_origin_info(vot.resources[0], "contact", "othercontact")

    with pytest.raises(ValueError, match="Unknown DataOrigin info name."):
        dataorigin.add_data_origin_info(vot.resources[0], "unexistingkey", "value")


def test_dataorigin_unsupported_input_error():
    table = Table.read(__generate_votable_test())

    with pytest.raises(TypeError, match="input vot_element type is not supported."):
        dataorigin.extract_data_origin(table)



# Licensed under a 3-clause BSD style license - see LICENSE.rst

import io

import pytest

from astropy.utils.compat.optional_deps import HAS_PLT

if HAS_PLT:
    from matplotlib.figure import Figure
    from matplotlib.units import ConversionError

import numpy as np

from astropy import units as u
from astropy.coordinates import Angle
from astropy.visualization.units import quantity_support


@pytest.mark.skipif(not HAS_PLT, reason="requires matplotlib")
def test_units():
    fig = Figure()
    ax = fig.add_subplot()

    with quantity_support():
        buff = io.BytesIO()

        ax.plot([1, 2, 3] * u.m, [3, 4, 5] * u.kg, label="label")
        ax.plot([105, 210, 315] * u.cm, [3050, 3025, 3010] * u.g)
        ax.legend()
        # Also test fill_between, which requires actual conversion to ndarray.
        ax.fill_between([1, 3] * u.m, [3, 5] * u.kg, [3050, 3010] * u.g)
        fig.savefig(buff, format="svg")

        assert ax.xaxis.get_units() == u.m
        assert ax.yaxis.get_units() == u.kg


@pytest.mark.skipif(not HAS_PLT, reason="requires matplotlib")
def test_units_decorator():
    @quantity_support()
    def run_test():
        fig = Figure()
        ax = fig.add_subplot()

        buff = io.BytesIO()

        ax.plot([1, 2, 3] * u.m, [3, 4, 5] * u.kg, label="label")
        ax.plot([105, 210, 315] * u.cm, [3050, 3025, 3010] * u.g)
        ax.legend()
        # Also test fill_between, which requires actual conversion to ndarray.
        ax.fill_between([1, 3] * u.m, [3, 5] * u.kg, [3050, 3010] * u.g)
        fig.savefig(buff, format="svg")

        assert ax.xaxis.get_units() == u.m
        assert ax.yaxis.get_units() == u.kg

    run_test()


@pytest.mark.skipif(not HAS_PLT, reason="requires matplotlib")
def test_units_errbarr():
    with quantity_support():
        x = [1, 2, 3] * u.s
        y = [1, 2, 3] * u.m
        yerr = [3, 2, 1] * u.cm

        fig = Figure()
        ax = fig.add_subplot()
        ax.errorbar(x, y, yerr=yerr)

        assert ax.xaxis.get_units() == u.s
        assert ax.yaxis.get_units() == u.m


@pytest.mark.skipif(not HAS_PLT, reason="requires matplotlib")
def test_incompatible_units():
    fig = Figure()
    ax = fig.add_subplot()

    with quantity_support():
        ax.plot([1, 2, 3] * u.m)
        with pytest.raises(ConversionError):
            ax.plot([105, 210, 315] * u.kg)


@pytest.mark.skipif(not HAS_PLT, reason="requires matplotlib")
def test_quantity_subclass():
    """Check that subclasses are recognized.

    Also see https://github.com/matplotlib/matplotlib/pull/13536
    """
    fig = Figure()
    ax = fig.add_subplot()

    with quantity_support():
        ax.scatter(Angle([1, 2, 3], u.deg), [3, 4, 5] * u.kg)
        ax.scatter([105, 210, 315] * u.arcsec, [3050, 3025, 3010] * u.g)
        ax.plot(Angle([105, 210, 315], u.arcsec), [3050, 3025, 3010] * u.g)

        assert ax.xaxis.get_units() == u.deg
        assert ax.yaxis.get_units() == u.kg


@pytest.mark.skipif(not HAS_PLT, reason="requires matplotlib")
def test_nested():
    with quantity_support():
        with quantity_support():
            fig = Figure()
            ax = fig.add_subplot(1, 1, 1)
            ax.scatter(Angle([1, 2, 3], u.deg), [3, 4, 5] * u.kg)

            assert ax.xaxis.get_units() == u.deg
            assert ax.yaxis.get_units() == u.kg

        fig = Figure()
        ax = fig.add_subplot(1, 1, 1)
        ax.scatter(Angle([1, 2, 3], u.arcsec), [3, 4, 5] * u.pc)

        assert ax.xaxis.get_units() == u.arcsec
        assert ax.yaxis.get_units() == u.pc


@pytest.mark.skipif(not HAS_PLT, reason="requires matplotlib")
def test_empty_hist():
    with quantity_support():
        fig = Figure()
        ax = fig.add_subplot(1, 1, 1)
        ax.hist([1, 2, 3, 4] * u.mmag, bins=100)
        # The second call results in an empty list being passed to the
        # unit converter in matplotlib >= 3.1
        ax.hist([] * u.mmag, bins=100)


@pytest.mark.skipif(not HAS_PLT, reason="requires matplotlib")
def test_radian_formatter():
    with quantity_support():
        fig = Figure()
        ax = fig.add_subplot()
        ax.plot([1, 2, 3], [1, 2, 3] * u.rad * np.pi)
        fig.canvas.draw()
        labels = [tl.get_text() for tl in ax.yaxis.get_ticklabels()]
        assert labels == ["π/2", "π", "3π/2", "2π", "5π/2", "3π", "7π/2"]


@pytest.mark.skipif(not HAS_PLT, reason="requires matplotlib")
def test_small_range():
    # see https://github.com/astropy/astropy/issues/13211
    y = [10.0, 10.25, 10.5, 10.75, 11.0, 11.25, 11.5, 11.75] * u.degree

    fig = Figure()
    ax = fig.add_subplot()

    with quantity_support():
        ax.plot(y)
        fig.canvas.draw()
    labels = [t.get_text() for t in ax.yaxis.get_ticklabels()]

    # check uniqueness of labels
    assert len(set(labels)) == len(labels)


@pytest.mark.skipif(not HAS_PLT, reason="requires matplotlib")
def test_override_axis_unit():
    fig = Figure()
    ax = fig.add_subplot()

    with quantity_support():
        ax.plot([1, 2, 3] * u.m)
        ax.yaxis.set_units(u.cm)
        fig.canvas.draw()
        assert ax.yaxis.get_units() == u.cm



"""
This file contains the code for Quantizing / Dequantizing floats.
"""

import numpy as np

from astropy.io.fits.hdu.base import BITPIX2DTYPE
from astropy.io.fits.hdu.compressed._compression import (
    quantize_double_c,
    quantize_float_c,
    unquantize_double_c,
    unquantize_float_c,
)

__all__ = ["Quantize"]


DITHER_METHODS = {
    "NONE": 0,
    "NO_DITHER": -1,
    "SUBTRACTIVE_DITHER_1": 1,
    "SUBTRACTIVE_DITHER_2": 2,
}


class QuantizationFailedException(Exception):
    pass


class Quantize:
    """
    Quantization of floating-point data following the FITS standard.
    """

    def __init__(
        self, *, row: int, dither_method: int, quantize_level: int, bitpix: int
    ):
        super().__init__()
        self.row = row
        # TODO: pass dither method as a string instead of int?
        self.quantize_level = quantize_level
        self.dither_method = dither_method
        self.bitpix = bitpix

    # NOTE: below we use decode_quantized and encode_quantized instead of
    # decode and encode as we need to break with the numcodec API and take/return
    # scale and zero in addition to quantized value. We should figure out how
    # to properly use the numcodec API for this use case.

    def decode_quantized(self, buf, scale, zero):
        """
        Unquantize data.

        Parameters
        ----------
        buf : bytes or array_like
            The buffer to unquantize.

        Returns
        -------
        np.ndarray
            The unquantized buffer.
        """
        qbytes = np.asarray(buf)
        qbytes = qbytes.astype(qbytes.dtype.newbyteorder("="))
        # TODO: figure out if we need to support null checking
        if self.dither_method == -1:
            # For NO_DITHER we should just use the scale and zero directly
            return qbytes * scale + zero
        if self.bitpix == -32:
            ubytes = unquantize_float_c(
                qbytes.tobytes(),
                self.row,
                qbytes.size,
                scale,
                zero,
                self.dither_method,
                0,
                0,
                0.0,
                qbytes.dtype.itemsize,
            )
        elif self.bitpix == -64:
            ubytes = unquantize_double_c(
                qbytes.tobytes(),
                self.row,
                qbytes.size,
                scale,
                zero,
                self.dither_method,
                0,
                0,
                0.0,
                qbytes.dtype.itemsize,
            )
        else:
            raise TypeError("bitpix should be one of -32 or -64")
        return np.frombuffer(ubytes, dtype=BITPIX2DTYPE[self.bitpix]).data

    def encode_quantized(self, buf):
        """
        Quantize data.

        Parameters
        ----------
        buf : bytes or array_like
            The buffer to quantize.

        Returns
        -------
        np.ndarray
            A buffer with quantized data.
        """
        uarray = np.asarray(buf)
        uarray = uarray.astype(uarray.dtype.newbyteorder("="))
        # TODO: figure out if we need to support null checking
        if uarray.dtype.itemsize == 4:
            qbytes, status, scale, zero = quantize_float_c(
                uarray.tobytes(),
                self.row,
                uarray.size,
                1,
                0,
                0,
                self.quantize_level,
                self.dither_method,
            )[:4]
        elif uarray.dtype.itemsize == 8:
            qbytes, status, scale, zero = quantize_double_c(
                uarray.tobytes(),
                self.row,
                uarray.size,
                1,
                0,
                0,
                self.quantize_level,
                self.dither_method,
            )[:4]
        if status == 0:
            raise QuantizationFailedException()
        else:
            return np.frombuffer(qbytes, dtype=np.int32), scale, zero



# Licensed under a 3-clause BSD style license - see LICENSE.rst

import os
import time

import numpy as np
import pytest

from astropy.io import fits
from astropy.utils.misc import _NOT_OVERWRITING_MSG_MATCH

from .conftest import FitsTestCase
from .test_table import comparerecords


class TestGroupsFunctions(FitsTestCase):
    def test_open(self):
        with fits.open(self.data("random_groups.fits")) as hdul:
            assert isinstance(hdul[0], fits.GroupsHDU)
            naxes = (3, 1, 128, 1, 1)
            parameters = ["UU", "VV", "WW", "BASELINE", "DATE"]
            info = [
                (
                    0,
                    "PRIMARY",
                    1,
                    "GroupsHDU",
                    147,
                    naxes,
                    "float32",
                    "3 Groups  5 Parameters",
                )
            ]
            assert hdul.info(output=False) == info

            ghdu = hdul[0]
            assert ghdu.parnames == parameters
            assert list(ghdu.data.dtype.names) == parameters + ["DATA"]

            assert isinstance(ghdu.data, fits.GroupData)
            # The data should be equal to the number of groups
            assert ghdu.header["GCOUNT"] == len(ghdu.data)
            assert ghdu.data.data.shape == (len(ghdu.data),) + naxes[::-1]
            assert ghdu.data.parnames == parameters

            assert isinstance(ghdu.data[0], fits.Group)
            assert len(ghdu.data[0]) == len(parameters) + 1
            assert ghdu.data[0].data.shape == naxes[::-1]
            assert ghdu.data[0].parnames == parameters

    def test_open_groups_in_update_mode(self):
        """
        Test that opening a file containing a groups HDU in update mode and
        then immediately closing it does not result in any unnecessary file
        modifications.

        Similar to
        test_image.TestImageFunctions.test_open_scaled_in_update_mode().
        """

        # Copy the original file before making any possible changes to it
        testfile = self.copy_file("random_groups.fits")
        mtime = os.stat(testfile).st_mtime

        time.sleep(1)

        fits.open(testfile, mode="update", memmap=False).close()

        # Ensure that no changes were made to the file merely by immediately
        # opening and closing it.
        assert mtime == os.stat(testfile).st_mtime

    def test_random_groups_data_update(self):
        """
        Regression test for https://github.com/astropy/astropy/issues/3730 and
        for https://github.com/spacetelescope/PyFITS/issues/102
        """

        testfile = self.copy_file("random_groups.fits")
        with fits.open(testfile, mode="update") as h:
            h[0].data["UU"] = 0.42

        with fits.open(testfile, mode="update") as h:
            assert np.all(h[0].data["UU"] == 0.42)

    def test_parnames_round_trip(self):
        """
        Regression test for https://aeon.stsci.edu/ssb/trac/pyfits/ticket/130

        Ensures that opening a random groups file in update mode or writing it
        to a new file does not cause any change to the parameter names.
        """

        # Because this test tries to update the random_groups.fits file, let's
        # make a copy of it first (so that the file doesn't actually get
        # modified in the off chance that the test fails
        testfile = self.copy_file("random_groups.fits")

        parameters = ["UU", "VV", "WW", "BASELINE", "DATE"]
        with fits.open(testfile, mode="update") as h:
            assert h[0].parnames == parameters
            h.flush()
        # Open again just in read-only mode to ensure the parnames didn't
        # change
        with fits.open(testfile) as h:
            assert h[0].parnames == parameters
            h.writeto(self.temp("test.fits"))

        with fits.open(self.temp("test.fits")) as h:
            assert h[0].parnames == parameters

    def test_groupdata_slice(self):
        """
        A simple test to ensure that slicing GroupData returns a new, smaller
        GroupData object, as is the case with a normal FITS_rec.  This is a
        regression test for an as-of-yet unreported issue where slicing
        GroupData returned a single Group record.
        """

        with fits.open(self.data("random_groups.fits")) as hdul:
            s = hdul[0].data[1:]
            assert isinstance(s, fits.GroupData)
            assert len(s) == 2
            assert hdul[0].data.parnames == s.parnames

    def test_group_slice(self):
        """
        Tests basic slicing a single group record.
        """

        # A very basic slice test
        with fits.open(self.data("random_groups.fits")) as hdul:
            g = hdul[0].data[0]
            s = g[2:4]
            assert len(s) == 2
            assert s[0] == g[2]
            assert s[-1] == g[-3]
            s = g[::-1]
            assert len(s) == 6
            assert (s[0] == g[-1]).all()
            assert s[-1] == g[0]
            s = g[::2]
            assert len(s) == 3
            assert s[0] == g[0]
            assert s[1] == g[2]
            assert s[2] == g[4]

    def test_create_groupdata(self, home_is_temp):
        """
        Basic test for creating GroupData from scratch.
        """

        imdata = np.arange(100.0).reshape((10, 1, 1, 2, 5))
        pdata1 = np.arange(10, dtype=np.float32) + 0.1
        pdata2 = 42.0
        x = fits.hdu.groups.GroupData(
            imdata, parnames=["abc", "xyz"], pardata=[pdata1, pdata2], bitpix=-32
        )
        assert x.parnames == ["abc", "xyz"]
        assert (x.par("abc") == pdata1).all()
        assert (x.par("xyz") == ([pdata2] * len(x))).all()
        assert (x.data == imdata).all()

        # Test putting the data into a GroupsHDU and round-tripping it
        ghdu = fits.GroupsHDU(data=x)
        assert ghdu.parnames == ["abc", "xyz"]
        assert ghdu.header["GCOUNT"] == 10

        filename = self.temp("test.fits")
        ghdu.writeto(filename)

        # Exercise the `overwrite` flag
        with pytest.raises(OSError, match=_NOT_OVERWRITING_MSG_MATCH):
            ghdu.writeto(filename, overwrite=False)
        ghdu.writeto(filename, overwrite=True)

        with fits.open(filename) as h:
            hdr = h[0].header
            assert hdr["GCOUNT"] == 10
            assert hdr["PCOUNT"] == 2
            assert hdr["NAXIS"] == 5
            assert hdr["NAXIS1"] == 0
            assert hdr["NAXIS2"] == 5
            assert hdr["NAXIS3"] == 2
            assert hdr["NAXIS4"] == 1
            assert hdr["NAXIS5"] == 1
            assert h[0].data.parnames == ["abc", "xyz"]
            assert comparerecords(h[0].data, x)

    def test_duplicate_parameter(self):
        """
        Tests support for multiple parameters of the same name, and ensures
        that the data in duplicate parameters are returned as a single summed
        value.
        """

        imdata = np.arange(100.0).reshape((10, 1, 1, 2, 5))
        pdata1 = np.arange(10, dtype=np.float32) + 1
        pdata2 = 42.0
        x = fits.hdu.groups.GroupData(
            imdata,
            parnames=["abc", "xyz", "abc"],
            pardata=[pdata1, pdata2, pdata1],
            bitpix=-32,
        )

        assert x.parnames == ["abc", "xyz", "abc"]
        assert (x.par("abc") == pdata1 * 2).all()
        assert x[0].par("abc") == 2

        # Test setting a parameter
        x[0].setpar(0, 2)
        assert x[0].par("abc") == 3
        pytest.raises(ValueError, x[0].setpar, "abc", 2)
        x[0].setpar("abc", (2, 3))
        assert x[0].par("abc") == 5
        assert x.par("abc")[0] == 5
        assert (x.par("abc")[1:] == pdata1[1:] * 2).all()

        # Test round-trip
        ghdu = fits.GroupsHDU(data=x)
        ghdu.writeto(self.temp("test.fits"))

        with fits.open(self.temp("test.fits")) as h:
            hdr = h[0].header
            assert hdr["PCOUNT"] == 3
            assert hdr["PTYPE1"] == "abc"
            assert hdr["PTYPE2"] == "xyz"
            assert hdr["PTYPE3"] == "abc"
            assert x.parnames == ["abc", "xyz", "abc"]
            assert x.dtype.names == ("abc", "xyz", "_abc", "DATA")
            assert x.par("abc")[0] == 5
            assert (x.par("abc")[1:] == pdata1[1:] * 2).all()

    def test_group_bad_naxis(self):
        """Test file without NAXIS1 keyword.
        Regression test for https://github.com/astropy/astropy/issues/9709
        """
        testfile = os.path.join("invalid", "group_invalid.fits")
        with fits.open(self.data(testfile)) as hdul:
            assert len(hdul) == 1
            assert hdul[0].header["GROUPS"]
            assert hdul[0].data is None

    def test_not_groups_file(self):
        hdu = fits.PrimaryHDU()
        hdu.header["GROUPS"] = (1, "not a groups HDU")
        hdu.writeto(self.temp("not_groups.fits"))
        with fits.open(self.temp("not_groups.fits")) as hdul:
            assert hdul[0].header["GROUPS"] == 1
            assert hdul[0].header.comments["GROUPS"] == "not a groups HDU"



# Licensed under a 3-clause BSD style license - see LICENSE.rst


import argparse
import copy
import logging
import sys
import time

from astropy import __version__, log

from .hub import SAMPHubServer

__all__ = ["hub_script"]


def hub_script(timeout=0):
    """
    This main function is executed by the ``samp_hub`` command line tool.
    """
    parser = argparse.ArgumentParser(prog="samp_hub " + __version__)
    # TODO: pass suggest_on_error as kwarg when PYTHON_LT_14 is dropped
    parser.suggest_on_error = True

    parser.add_argument(
        "-k", "--secret", dest="secret", metavar="CODE", help="custom secret code."
    )

    parser.add_argument(
        "-d", "--addr", dest="addr", metavar="ADDR", help="listening address (or IP)."
    )

    parser.add_argument(
        "-p",
        "--port",
        dest="port",
        metavar="PORT",
        type=int,
        help="listening port number.",
    )

    parser.add_argument(
        "-f", "--lockfile", dest="lockfile", metavar="FILE", help="custom lockfile."
    )

    parser.add_argument(
        "-w",
        "--no-web-profile",
        dest="web_profile",
        action="store_false",
        help="run the Hub disabling the Web Profile.",
        default=True,
    )

    parser.add_argument(
        "-P",
        "--pool-size",
        dest="pool_size",
        metavar="SIZE",
        type=int,
        help="the socket connections pool size.",
        default=20,
    )

    timeout_group = parser.add_argument_group(
        "Timeout group",
        "Special options to setup hub and client timeouts."
        "It contains a set of special options that allows to set up the Hub and "
        "clients inactivity timeouts, that is the Hub or client inactivity time "
        "interval after which the Hub shuts down or unregisters the client. "
        "Notification of samp.hub.disconnect MType is sent to the clients "
        "forcibly unregistered for timeout expiration.",
    )

    timeout_group.add_argument(
        "-t",
        "--timeout",
        dest="timeout",
        metavar="SECONDS",
        help=(
            "set the Hub inactivity timeout in SECONDS. By default it "
            "is set to 0, that is the Hub never expires."
        ),
        type=int,
        default=0,
    )

    timeout_group.add_argument(
        "-c",
        "--client-timeout",
        dest="client_timeout",
        metavar="SECONDS",
        help=(
            "set the client inactivity timeout in SECONDS. By default it "
            "is set to 0, that is the client never expires."
        ),
        type=int,
        default=0,
    )

    parser.add_argument_group(timeout_group)

    log_group = parser.add_argument_group(
        "Logging options",
        "Additional options which allow to customize the logging output. By "
        "default the SAMP Hub uses the standard output and standard error "
        "devices to print out INFO level logging messages. Using the options "
        "here below it is possible to modify the logging level and also "
        "specify the output files where redirect the logging messages.",
    )

    # valid logging levels as strings, reverse ordered by level.
    # the default 'INFO' is assumed to exist (...)
    log_levels = sorted(d := logging.getLevelNamesMapping(), key=d.get, reverse=True)
    log_group.add_argument(
        "-L",
        "--log-level",
        dest="loglevel",
        metavar="LEVEL",
        help=f"set the Hub instance log level ({', '.join(log_levels)}).",
        type=str,
        choices=log_levels,
        default="INFO",
    )
    del log_levels

    log_group.add_argument(
        "-O",
        "--log-output",
        dest="logout",
        metavar="FILE",
        help="set the output file for the log messages.",
        default="",
    )

    parser.add_argument_group(log_group)

    adv_group = parser.add_argument_group(
        "Advanced group",
        "Advanced options addressed to facilitate administrative tasks and "
        "allow new non-standard Hub behaviors. In particular the --label "
        "options is used to assign a value to hub.label token and is used to "
        "assign a name to the Hub instance. "
        "The very special --multi option allows to start a Hub in multi-instance mode. "
        "Multi-instance mode is a non-standard Hub behavior that enables "
        "multiple contemporaneous running Hubs. Multi-instance hubs place "
        "their non-standard lock-files within the <home directory>/.samp-1 "
        "directory naming them making use of the format: "
        "samp-hub-<PID>-<ID>, where PID is the Hub process ID while ID is an "
        "internal ID (integer).",
    )

    adv_group.add_argument(
        "-l",
        "--label",
        dest="label",
        metavar="LABEL",
        help="assign a LABEL to the Hub.",
        default="",
    )

    adv_group.add_argument(
        "-m",
        "--multi",
        dest="mode",
        help=(
            "run the Hub in multi-instance mode generating a custom "
            "lockfile with a random name."
        ),
        action="store_const",
        const="multiple",
        default="single",
    )

    parser.add_argument_group(adv_group)

    options = parser.parse_args()

    try:
        log.setLevel(options.loglevel)

        if options.logout != "":
            context = log.log_to_file(options.logout)
        else:

            class dummy_context:
                def __enter__(self):
                    pass

                def __exit__(self, exc_type, exc_value, traceback):
                    pass

            context = dummy_context()

        with context:
            args = copy.deepcopy(options.__dict__)
            del args["loglevel"]
            del args["logout"]

            hub = SAMPHubServer(**args)
            hub.start(False)

            if not timeout:
                while hub.is_running:
                    time.sleep(0.01)
            else:
                time.sleep(timeout)
                hub.stop()

    except KeyboardInterrupt:
        try:
            hub.stop()
        except NameError:
            pass
    except OSError as e:
        print(f"[SAMP] Error: I/O error({e.errno}): {e.strerror}")
        sys.exit(1)
    except SystemExit:
        pass



# Licensed under a 3-clause BSD style license - see LICENSE.rst


from numpy import exp
from numpy.typing import ArrayLike

from astropy.cosmology._src.core import dataclass_decorator
from astropy.cosmology._src.parameter import Parameter
from astropy.cosmology._src.typing import FArray
from astropy.cosmology._src.utils import aszarr, deprecated_keywords
from astropy.units import Quantity

from . import scalar_inv_efuncs
from .base import FLRW, FlatFLRWMixin

__all__ = ["Flatw0waCDM", "w0waCDM"]

__doctest_requires__ = {"*": ["scipy"]}


@dataclass_decorator
class w0waCDM(FLRW):
    r"""FLRW cosmology with a CPL dark energy EoS and curvature.

    The equation for the dark energy equation of state (EoS) uses the
    CPL form as described in Chevallier & Polarski [1]_ and Linder [2]_:
    :math:`w(z) = w_0 + w_a (1-a) = w_0 + w_a z / (1+z)`.

    Parameters
    ----------
    H0 : float or scalar quantity-like ['frequency']
        Hubble constant at z = 0. If a float, must be in [km/sec/Mpc].

    Om0 : float
        Omega matter: density of non-relativistic matter in units of the
        critical density at z=0.

    Ode0 : float
        Omega dark energy: density of dark energy in units of the critical
        density at z=0.

    w0 : float, optional
        Dark energy equation of state at z=0 (a=1). This is pressure/density
        for dark energy in units where c=1.

    wa : float, optional
        Negative derivative of the dark energy equation of state with respect
        to the scale factor. A cosmological constant has w0=-1.0 and wa=0.0.

    Tcmb0 : float or scalar quantity-like ['temperature'], optional
        Temperature of the CMB z=0. If a float, must be in [K]. Default: 0 [K].
        Setting this to zero will turn off both photons and neutrinos
        (even massive ones).

    Neff : float, optional
        Effective number of Neutrino species. Default 3.04.

    m_nu : quantity-like ['energy', 'mass'] or array-like, optional
        Mass of each neutrino species in [eV] (mass-energy equivalency enabled).
        If this is a scalar Quantity, then all neutrino species are assumed to
        have that mass. Otherwise, the mass of each species. The actual number
        of neutrino species (and hence the number of elements of m_nu if it is
        not scalar) must be the floor of Neff. Typically this means you should
        provide three neutrino masses unless you are considering something like
        a sterile neutrino.

    Ob0 : float or None, optional
        Omega baryons: density of baryonic matter in units of the critical
        density at z=0.  If this is set to None (the default), any computation
        that requires its value will raise an exception.

    name : str or None (optional, keyword-only)
        Name for this cosmological object.

    meta : mapping or None (optional, keyword-only)
        Metadata for the cosmology, e.g., a reference.

    Examples
    --------
    >>> from astropy.cosmology import w0waCDM
    >>> cosmo = w0waCDM(H0=70, Om0=0.3, Ode0=0.7, w0=-0.9, wa=0.2)

    The comoving distance in Mpc at redshift z:

    >>> z = 0.5
    >>> dc = cosmo.comoving_distance(z)

    References
    ----------
    .. [1] Chevallier, M., & Polarski, D. (2001). Accelerating Universes with
           Scaling Dark Matter. International Journal of Modern Physics D,
           10(2), 213-223.
    .. [2] Linder, E. (2003). Exploring the Expansion History of the
           Universe. Phys. Rev. Lett., 90, 091301.
    """

    w0: Parameter = Parameter(
        default=-1.0, doc="Dark energy equation of state at z=0.", fvalidate="float"
    )
    wa: Parameter = Parameter(
        default=0.0,
        doc="Negative derivative of dark energy equation of state w.r.t. a.",
        fvalidate="float",
    )

    def __post_init__(self) -> None:
        super().__post_init__()

        # Please see :ref:`astropy-cosmology-fast-integrals` for discussion
        # about what is being done here.
        if self.Tcmb0.value == 0:
            inv_efunc_scalar = scalar_inv_efuncs.w0wacdm_inv_efunc_norel
            inv_efunc_scalar_args = (
                self.Om0,
                self.Ode0,
                self.Ok0,
                self.w0,
                self.wa,
            )
        elif not self._nu_info.has_massive_nu:
            inv_efunc_scalar = scalar_inv_efuncs.w0wacdm_inv_efunc_nomnu
            inv_efunc_scalar_args = (
                self.Om0,
                self.Ode0,
                self.Ok0,
                self.Ogamma0 + self.Onu0,
                self.w0,
                self.wa,
            )
        else:
            inv_efunc_scalar = scalar_inv_efuncs.w0wacdm_inv_efunc
            inv_efunc_scalar_args = (
                self.Om0,
                self.Ode0,
                self.Ok0,
                self.Ogamma0,
                self._nu_info.neff_per_nu,
                self._nu_info.n_massless_nu,
                self._nu_info.nu_y_list,
                self.w0,
                self.wa,
            )
        object.__setattr__(self, "_inv_efunc_scalar", inv_efunc_scalar)
        object.__setattr__(self, "_inv_efunc_scalar_args", inv_efunc_scalar_args)

    @deprecated_keywords("z", since="7.0")
    def w(self, z: Quantity | ArrayLike) -> FArray:
        r"""Returns dark energy equation of state at redshift ``z``.

        Parameters
        ----------
        z : Quantity-like ['redshift'] or array-like
            Input redshift.

            .. versionchanged:: 7.0
                Passing z as a keyword argument is deprecated.

        Returns
        -------
        w : ndarray
            The dark energy equation of state

        Notes
        -----
        The dark energy equation of state is defined as
        :math:`w(z) = P(z)/\rho(z)`, where :math:`P(z)` is the pressure at
        redshift z and :math:`\rho(z)` is the density at redshift z, both in
        units where c=1. Here this is
        :math:`w(z) = w_0 + w_a (1 - a) = w_0 + w_a \frac{z}{1+z}`.
        """
        z = aszarr(z)
        return self.w0 + self.wa * z / (z + 1.0)

    @deprecated_keywords("z", since="7.0")
    def de_density_scale(self, z: Quantity | ArrayLike) -> FArray:
        r"""Evaluates the redshift dependence of the dark energy density.

        Parameters
        ----------
        z : Quantity-like ['redshift'] or array-like, positional-only
            Input redshift.

            .. versionchanged:: 7.0
                Passing z as a keyword argument is deprecated.

        Returns
        -------
        I : ndarray
            The scaling of the energy density of dark energy with redshift.

        Notes
        -----
        The scaling factor, I, is defined by :math:`\rho(z) = \rho_0 I`,
        and in this case is given by

        .. math::

           I = \left(1 + z\right)^{3 \left(1 + w_0 + w_a\right)}
                     \exp \left(-3 w_a \frac{z}{1+z}\right)
        """
        z = aszarr(z)
        zp1 = z + 1.0  # (converts z [unit] -> z [dimensionless])
        return zp1 ** (3 * (1 + self.w0 + self.wa)) * exp(-3 * self.wa * z / zp1)


@dataclass_decorator
class Flatw0waCDM(FlatFLRWMixin, w0waCDM):
    """FLRW cosmology with a CPL dark energy EoS and no curvature.

    The equation for the dark energy equation of state (EoS) uses the CPL form as
    described in Chevallier & Polarski [1]_ and Linder [2]_:
    :math:`w(z) = w_0 + w_a (1-a) = w_0 + w_a z / (1+z)`.

    Parameters
    ----------
    H0 : float or scalar quantity-like ['frequency']
        Hubble constant at z = 0. If a float, must be in [km/sec/Mpc].

    Om0 : float
        Omega matter: density of non-relativistic matter in units of the
        critical density at z=0.

    w0 : float, optional
        Dark energy equation of state at z=0 (a=1). This is pressure/density
        for dark energy in units where c=1.

    wa : float, optional
        Negative derivative of the dark energy equation of state with respect
        to the scale factor. A cosmological constant has w0=-1.0 and wa=0.0.

    Tcmb0 : float or scalar quantity-like ['temperature'], optional
        Temperature of the CMB z=0. If a float, must be in [K]. Default: 0 [K].
        Setting this to zero will turn off both photons and neutrinos
        (even massive ones).

    Neff : float, optional
        Effective number of Neutrino species. Default 3.04.

    m_nu : quantity-like ['energy', 'mass'] or array-like, optional
        Mass of each neutrino species in [eV] (mass-energy equivalency enabled).
        If this is a scalar Quantity, then all neutrino species are assumed to
        have that mass. Otherwise, the mass of each species. The actual number
        of neutrino species (and hence the number of elements of m_nu if it is
        not scalar) must be the floor of Neff. Typically this means you should
        provide three neutrino masses unless you are considering something like
        a sterile neutrino.

    Ob0 : float or None, optional
        Omega baryons: density of baryonic matter in units of the critical
        density at z=0.  If this is set to None (the default), any computation
        that requires its value will raise an exception.

    name : str or None (optional, keyword-only)
        Name for this cosmological object.

    meta : mapping or None (optional, keyword-only)
        Metadata for the cosmology, e.g., a reference.

    Examples
    --------
    >>> from astropy.cosmology import Flatw0waCDM
    >>> cosmo = Flatw0waCDM(H0=70, Om0=0.3, w0=-0.9, wa=0.2)

    The comoving distance in Mpc at redshift z:

    >>> z = 0.5
    >>> dc = cosmo.comoving_distance(z)

    To get an equivalent cosmology, but of type `astropy.cosmology.w0waCDM`,
    use :attr:`astropy.cosmology.FlatFLRWMixin.nonflat`.

    >>> print(cosmo.nonflat)
    w0waCDM(H0=70.0 km / (Mpc s), Om0=0.3, Ode0=0.7, ...

    References
    ----------
    .. [1] Chevallier, M., & Polarski, D. (2001). Accelerating Universes with
           Scaling Dark Matter. International Journal of Modern Physics D,
           10(2), 213-223.
    .. [2] Linder, E. (2003). Exploring the Expansion History of the
           Universe. Phys. Rev. Lett., 90, 091301.
    """

    def __post_init__(self) -> None:
        super().__post_init__()

        # Please see :ref:`astropy-cosmology-fast-integrals` for discussion
        # about what is being done here.
        if self.Tcmb0.value == 0:
            inv_efunc_scalar = scalar_inv_efuncs.fw0wacdm_inv_efunc_norel
            inv_efunc_scalar_args = (self.Om0, self.Ode0, self.w0, self.wa)
        elif not self._nu_info.has_massive_nu:
            inv_efunc_scalar = scalar_inv_efuncs.fw0wacdm_inv_efunc_nomnu
            inv_efunc_scalar_args = (
                self.Om0,
                self.Ode0,
                self.Ogamma0 + self.Onu0,
                self.w0,
                self.wa,
            )
        else:
            inv_efunc_scalar = scalar_inv_efuncs.fw0wacdm_inv_efunc
            inv_efunc_scalar_args = (
                self.Om0,
                self.Ode0,
                self.Ogamma0,
                self._nu_info.neff_per_nu,
                self._nu_info.n_massless_nu,
                self._nu_info.nu_y_list,
                self.w0,
                self.wa,
            )
        object.__setattr__(self, "_inv_efunc_scalar", inv_efunc_scalar)
        object.__setattr__(self, "_inv_efunc_scalar_args", inv_efunc_scalar_args)



# Licensed under a 3-clause BSD style license - see LICENSE.rst

"""Tests for :mod:`astropy.cosmology.comparison`"""

import re

import numpy as np
import pytest

from astropy.cosmology import Cosmology, FlatCosmologyMixin, Planck18, cosmology_equal
from astropy.cosmology._src.funcs.comparison import (
    _CANT_BROADCAST,
    _cosmology_not_equal,
    _CosmologyWrapper,
    _parse_format,
    _parse_formats,
)
from astropy.cosmology._src.tests.io.base import ToFromTestMixinBase
from astropy.cosmology.io import convert_registry


class ComparisonFunctionTestBase(ToFromTestMixinBase):
    """Tests for cosmology comparison functions.

    This class inherits from
    `astropy.cosmology._src.tests.io.base.ToFromTestMixinBase` because the cosmology
    comparison functions all have a kwarg ``format`` that allow the arguments to
    be converted to a |Cosmology| using the ``to_format`` architecture.

    This class will not be directly called by :mod:`pytest` since its name does
    not begin with ``Test``. To activate the contained tests this class must be
    inherited in a subclass.
    """

    @pytest.fixture(scope="class")
    def cosmo(self):
        return Planck18

    @pytest.fixture(scope="class")
    def cosmo_eqvxflat(self, cosmo):
        if isinstance(cosmo, FlatCosmologyMixin):
            return cosmo.nonflat

        pytest.skip(
            "cosmology is not flat, so does not have an equivalent non-flat cosmology."
        )

    @pytest.fixture(
        scope="class",
        params=sorted(
            {k for k, _ in convert_registry._readers.keys()} - {"astropy.cosmology"}
        ),
    )
    def format(self, request):
        return request.param

    @pytest.fixture(scope="class")
    def xfail_cant_autoidentify(self, format):
        """`pytest.fixture` form of method ``can_autoidentify`."""
        if not self.can_autodentify(format):
            pytest.xfail("cannot autoidentify")

    @pytest.fixture(scope="class")
    def converted(self, to_format, format):
        if format == "astropy.model":  # special case Model
            return to_format(format, method="comoving_distance")
        return to_format(format)

    @pytest.fixture(scope="class")
    def pert_cosmo(self, cosmo):
        # change one parameter
        p, v = next(iter(cosmo.parameters.items()))
        return cosmo.clone(
            **{p: v * 1.0001 if v != 0 else 0.001 * getattr(v, "unit", 1)}
        )

    @pytest.fixture(scope="class")
    def pert_cosmo_eqvxflat(self, pert_cosmo):
        if isinstance(pert_cosmo, FlatCosmologyMixin):
            return pert_cosmo.nonflat

        pytest.skip(
            "cosmology is not flat, so does not have an equivalent non-flat cosmology."
        )

    @pytest.fixture(scope="class")
    def pert_converted(self, pert_cosmo, format):
        if format == "astropy.model":  # special case Model
            return pert_cosmo.to_format(format, method="comoving_distance")
        return pert_cosmo.to_format(format)


class Test_parse_format(ComparisonFunctionTestBase):
    """Test functions ``_parse_format``."""

    @pytest.fixture(scope="class")
    def converted(self, to_format, format):
        if format == "astropy.model":  # special case Model
            return to_format(format, method="comoving_distance")

        converted = to_format(format)

        # Some raise a segfault! TODO: figure out why
        if isinstance(converted, _CANT_BROADCAST):
            converted = _CosmologyWrapper(converted)

        return converted

    # ========================================================================

    def test_shortcut(self, cosmo):
        """Test the already-a-cosmology shortcut."""
        # A Cosmology
        for fmt in (None, True, False, "astropy.cosmology"):
            assert _parse_format(cosmo, fmt) is cosmo, f"{fmt} failed"

        # A Cosmology, but improperly formatted
        # see ``test_parse_format_error_wrong_format``.

    def test_convert(self, converted, format, cosmo):
        """Test converting a cosmology-like object"""
        out = _parse_format(converted, format)

        assert isinstance(out, Cosmology)
        assert out == cosmo

    def test_parse_format_error_wrong_format(self, cosmo):
        """
        Test ``_parse_format`` errors when given a Cosmology object and format
        is not compatible.
        """
        with pytest.raises(
            ValueError, match=re.escape("for parsing a Cosmology, 'format'")
        ):
            _parse_format(cosmo, "mapping")

    def test_parse_format_error_noncosmology_cant_convert(self):
        """
        Test ``_parse_format`` errors when given a non-Cosmology object
        and format is `False`.
        """
        notacosmo = object()

        with pytest.raises(TypeError, match=re.escape("if 'format' is False")):
            _parse_format(notacosmo, False)

    def test_parse_format_vectorized(self, cosmo, format, converted):
        # vectorized on cosmos
        out = _parse_format([cosmo, cosmo], None)
        assert len(out) == 2
        assert np.all(out == cosmo)

        # vectorized on formats
        out = _parse_format(cosmo, [None, None])
        assert len(out) == 2
        assert np.all(out == cosmo)

        # more complex broadcast
        out = _parse_format(
            [[cosmo, converted], [converted, cosmo]], [[None, format], [format, None]]
        )
        assert out.shape == (2, 2)
        assert np.all(out == cosmo)

    def test_parse_formats_vectorized(self, cosmo):
        # vectorized on cosmos
        out = _parse_formats(cosmo, cosmo, format=None)
        assert len(out) == 2
        assert np.all(out == cosmo)

        # does NOT vectorize on formats
        with pytest.raises(ValueError, match="operands could not be broadcast"):
            _parse_formats(cosmo, format=[None, None])


class Test_cosmology_equal(ComparisonFunctionTestBase):
    """Test :func:`astropy.cosmology.comparison.cosmology_equal`"""

    def test_cosmology_equal_simple(self, cosmo, pert_cosmo):
        # equality
        assert cosmology_equal(cosmo, cosmo) is True

        # not equal to perturbed cosmology
        assert cosmology_equal(cosmo, pert_cosmo) is False

    def test_cosmology_equal_equivalent(
        self, cosmo, cosmo_eqvxflat, pert_cosmo, pert_cosmo_eqvxflat
    ):
        # now need to check equivalent, but not equal, cosmologies.
        assert cosmology_equal(cosmo, cosmo_eqvxflat, allow_equivalent=True) is True
        assert cosmology_equal(cosmo, cosmo_eqvxflat, allow_equivalent=False) is False

        assert (
            cosmology_equal(pert_cosmo, pert_cosmo_eqvxflat, allow_equivalent=True)
            is True
        )
        assert (
            cosmology_equal(pert_cosmo, pert_cosmo_eqvxflat, allow_equivalent=False)
            is False
        )

    def test_cosmology_equal_too_many_cosmo(self, cosmo):
        with pytest.raises(
            TypeError, match="cosmology_equal takes 2 positional arguments"
        ):
            cosmology_equal(cosmo, cosmo, cosmo)

    def test_cosmology_equal_format_error(self, cosmo, converted):
        # Not converting `converted`
        with pytest.raises(TypeError):
            cosmology_equal(cosmo, converted)

        with pytest.raises(TypeError):
            cosmology_equal(cosmo, converted, format=False)

    def test_cosmology_equal_format_auto(
        self, cosmo, converted, xfail_cant_autoidentify
    ):
        # These tests only run if the format can autoidentify.
        assert cosmology_equal(cosmo, converted, format=None) is True
        assert cosmology_equal(cosmo, converted, format=True) is True

    def test_cosmology_equal_format_specify(
        self, cosmo, format, converted, pert_converted
    ):
        # equality
        assert cosmology_equal(cosmo, converted, format=[None, format]) is True
        assert cosmology_equal(converted, cosmo, format=[format, None]) is True

        # non-equality
        assert cosmology_equal(cosmo, pert_converted, format=[None, format]) is False

    def test_cosmology_equal_equivalent_format_specify(
        self, cosmo, format, converted, cosmo_eqvxflat
    ):
        # specifying the format
        assert (
            cosmology_equal(
                cosmo_eqvxflat, converted, format=[None, format], allow_equivalent=True
            )
            is True
        )
        assert (
            cosmology_equal(
                converted, cosmo_eqvxflat, format=[format, None], allow_equivalent=True
            )
            is True
        )


class Test_cosmology_not_equal(ComparisonFunctionTestBase):
    """Test :func:`astropy.cosmology.comparison._cosmology_not_equal`"""

    def test_cosmology_not_equal_simple(self, cosmo, pert_cosmo):
        # equality
        assert _cosmology_not_equal(cosmo, cosmo) is False

        # not equal to perturbed cosmology
        assert _cosmology_not_equal(cosmo, pert_cosmo) is True

    def test_cosmology_not_equal_too_many_cosmo(self, cosmo):
        with pytest.raises(TypeError, match="_cosmology_not_equal takes 2 positional"):
            _cosmology_not_equal(cosmo, cosmo, cosmo)

    def test_cosmology_not_equal_equivalent(
        self, cosmo, cosmo_eqvxflat, pert_cosmo, pert_cosmo_eqvxflat
    ):
        # now need to check equivalent, but not equal, cosmologies.
        assert (
            _cosmology_not_equal(cosmo, cosmo_eqvxflat, allow_equivalent=False) is True
        )
        assert (
            _cosmology_not_equal(cosmo, cosmo_eqvxflat, allow_equivalent=True) is False
        )

        assert (
            _cosmology_not_equal(
                pert_cosmo, pert_cosmo_eqvxflat, allow_equivalent=False
            )
            is True
        )
        assert (
            _cosmology_not_equal(pert_cosmo, pert_cosmo_eqvxflat, allow_equivalent=True)
            is False
        )

    def test_cosmology_not_equal_format_error(self, cosmo, converted):
        # Not converting `converted`
        with pytest.raises(TypeError):
            _cosmology_not_equal(cosmo, converted)

        with pytest.raises(TypeError):
            _cosmology_not_equal(cosmo, converted, format=False)

    def test_cosmology_not_equal_format_auto(
        self, cosmo, pert_converted, xfail_cant_autoidentify
    ):
        assert _cosmology_not_equal(cosmo, pert_converted, format=None) is True
        assert _cosmology_not_equal(cosmo, pert_converted, format=True) is True

    def test_cosmology_not_equal_format_specify(
        self, cosmo, format, converted, pert_converted
    ):
        # specifying the format
        assert (
            _cosmology_not_equal(cosmo, pert_converted, format=[None, format]) is True
        )
        assert (
            _cosmology_not_equal(pert_converted, cosmo, format=[format, None]) is True
        )

        # equality
        assert _cosmology_not_equal(cosmo, converted, format=[None, format]) is False

    def test_cosmology_not_equal_equivalent_format_specify(
        self, cosmo, format, converted, cosmo_eqvxflat
    ):
        # specifying the format
        assert (
            _cosmology_not_equal(
                cosmo_eqvxflat, converted, format=[None, format], allow_equivalent=False
            )
            is True
        )
        assert (
            _cosmology_not_equal(
                cosmo_eqvxflat, converted, format=[None, format], allow_equivalent=True
            )
            is False
        )

        assert (
            _cosmology_not_equal(
                converted, cosmo_eqvxflat, format=[format, None], allow_equivalent=True
            )
            is False
        )
